/*
barcode.cpp
Copyright (C)2001  Al Kingsley


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY.

*/

#include <fcntl.h>
#include <unistd.h>

#include "pos.h"
#include "bhbarcode.h"

namespace bhpos{
   
namespace helper{    

CBarcodeHelper::CBarcodeHelper()
{
	data=NULL;
	type=UNKNOWN;
}
CBarcodeHelper::CBarcodeHelper(const char *b)
{
	if(b){
	   if(strlen(b) > 13){
	   	data=new char[14];
	   }
	   else{
	      data=new char[strlen(b)+1];
	   }
		if(data==NULL) return;
		
		memcpy(data, b, 13);
		data[13]=0;
	}
}

CBarcodeHelper::~CBarcodeHelper()
{
	if(data!=NULL) delete data;
	data=NULL;
}

int CBarcodeHelper::Set(const char *b)
{
	if(b){
	   if(data!=NULL) delete data;
		if(strlen(b) > 13){
	   	data=new char[14];
	   }
	   else{
	      data=new char[strlen(b)+1];
	   }
		if(data==NULL) return 1;
		
		memcpy(data, b, 13);
		data[13]=0;
	}
	return 0;
}
int CBarcodeHelper::GetNumber(char c)
{
	char buff[2]={c, 0};
	if(c < 48 || c > 57) return -1;
	return atoi(buff);
}

int CBarcodeHelper::LunBun()
{
	int i, value=0; 
	int len=0;
	
	if(data==NULL) return -1;
	len=strlen(data);
	
	int uwstart=len%2;
	
	for(i=0; i<=len; i++){
		if(uwstart){	
			value+=GetNumber(data[i]);
			uwstart=1;
		}
		else{	// // even num digits - weight on odd dig
			value+=GetNumber(data[i]) * 3;
			uwstart=1;
		}
	}
	return (10-(value%10)); 
}

int CBarcodeHelper::Size()
{
	if(data) return strlen(data);
	return 0;
}

bool CBarcodeHelper::IsValid()
{
	int len, odd=1;
		
	if(data==NULL) return false;
	int cdigit=LunBun();
	if(cdigit==-1) return false;
		
	len=strlen(data)-1;
	if(GetNumber(data[len])!=cdigit) return false;
		
	return true;
}
int CBarcodeHelper::FormatTo(BarcodeType t, unsigned int flags, int size)
{
	int 	len, nlen=13, slen=size;
	bool 	bAlpha=false;
	char	*newdata=NULL;
	
	if(data==NULL) return 1;
	
	len=strlen(data);
	
	switch(t){
	case UNKNOWN: return 2;
	
	// unsupported
	case CODABLOCK:
	case CODE1:
	case CODE16K:
	case CODE11:
	case CODE49:
	case CODE93:
	case FIM:
	case I2OF5:
	case MSI:
	case NI2OF5:
	case PHARMA:
	case POSTNET: 
		return 3;
	
	// to be supported
	case CODE128:
	case CODE39:
	case CODE2OF5:
	case CODABAR:
	case EAN128:
		return 4;
	
	case JAN8: 
	case EAN8:
		nlen=8;
		break;
	case JAN8_2:
	case EAN8_2:
	case ISBN:
		nlen=10;
		break;	
	case JAN8_5:
	case EAN8_5:
	case JAN13:
	case EAN13:
		nlen=13;
		break;
	case JAN13_2:
	case EAN13_2:
	case EANISBN:	// preceded by "978"
		nlen=15;
		break;
	case JAN13_5:
	case EAN13_5:
		case ISBN_5:
		nlen=18;
		break;
		
	case UPCA:
	case UPCE_5:
		nlen=12;
		break;
	case UPCA_2:
		nlen=14;
		break;
	case UPCA_5:
		nlen=17;
		break;
	case UPCE:
		nlen=8;
		break;
	case UPCE_2:
		nlen=10;
		break;	
	
	case BH_SHORTCODE6:
		nlen=6;
		break;
	}
		
	if(flags==BCF_EXACTLENGTH){
		return 0;
	}
	else if(flags==BCF_ZEROPADDLEFT){
		if(len==0 || size < len) return 0; 
		newdata=new char[size+1];
		if(newdata==NULL) return 5;		
		memset(newdata, '0', size - 1);
		strcpy(&newdata[size-strlen(data)], data);
		// allow for custom codes in UPCA to EAN13
		if(len==12 && nlen==13){
			if(data[0]=='2' || data[0]=='4'){
				newdata[size-strlen(data)-1]='2';
			}
		}
	}
	else if(flags==BCF_ZEROPADDRIGHT){
		if(len==0 || size < len) return 0;
		newdata=new char[size+1];
		if(newdata==NULL) return 6;
		memset(newdata, '0', size-1);
		memcpy(newdata, data, strlen(data));
		newdata[size-1]=0;
	}
	else{
		return 7;
	}
	delete data;
	data=NULL;
	data=newdata;
	
	return 0;
}
const char *CBarcodeHelper::Get()
{
	if(data) return data;
	return NULL;
}

}

//////////////////////////////////////////////////////////
//
//
//
//////////////////////////////////////////////////////////

CBarcode::CBarcode(BhDevType type) : CBhDevice(type)
{
}

CBarcode::~CBarcode()
{
	Close();
}

int CBarcode::OpenScanner(const char *dev, BhDevNotifyType nt)
{
   return Open(dev, nt);
}

/*
void CBarcode::Input(int len)
{
	
	int cdigit;

	if(len==8){
		memcpy(&indata[5], &indata[0], 8);
		indata[0]=0;
		indata[1]=0;
		indata[2]=0;
		indata[3]=0;
		indata[4]=0;
	}

	if(len==12){
		memcpy(&indata[1], &indata[0], 12);
		indata[0]=0;
	}

	cdigit=indata[0] + indata[2] + indata[4] + indata[6] + indata[8] + indata[10] + indata[12] +
		(indata[1] * 3) + (indata[3] * 3) + (indata[5] * 3) + (indata[7] * 3) + (indata[9] * 3) + 
		(indata[11] * 3);
		
	cdigit%=10;
	if(atoi(&indata[12])!=cdigit){
		CBhDevice::Dispatch(BHDEV_READ_ERROR);
		return;
	}
	CBhDevice::Dispatch(BHDEV_READ_DATA);
}
*/
}
