/*
bhcashd.cpp - a cash drawer shared library wrapper.
Copyright (C) 2001  Al Kingsley


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Don't write to me at: kingsley@bananahead.com 

	Part of:
	BhPos, Copyright (C) 2001-2005 Al Kingsley
	BhPos comes with ABSOLUTELY NO WARRANTY. 
*/
#include <cstring>
#include <unistd.h>
#include <fcntl.h>
#include <cstdio>
#include <cstddef>
#include <cstring>
#include <cstdlib>

#include "pos.h"
#include "bhcashd.h"

namespace bhpos{ 
	
CCashD::CCashD(BhDevType type) : CBhDevice(type)
{
	status=0;
	device=-1;
}

CCashD::~CCashD()
{
	if(device!=-1) close(device);
}

bool CCashD::SetDevice(char *dev)
{
	device=open(dev, O_RDONLY);
	if(device==-1) return true;
	return false;
}

void CCashD::SetOption(CashdrawerOption o, void *d)
{
   switch(o){
   case CASHD_OPTION_FIRESTRING:
      firestring=(char *) d;
      break;
   }
}

int CCashD::Fire()
{
	char state=1;
	
	if(device==-1) return -2;
	return write(device, &state, 1);
}

int CCashD::IsOpen()
{
	int 	size;
	char	state;
	
	if(device==-1) return -2;
	
	size=read(device, &state, 1);	
	if(size==1){
		if(state==CASHD_STATUS_CLOSED) return CASHD_STATUS_CLOSED;
		else if(state==CASHD_STATUS_OPEN) return CASHD_STATUS_OPEN;
		else return -1;
	}
	return -1;
}
/*
int CMfPosEngine::CashDrawerTimer()
{
   
   static int beep=0;
   Gdk::Color col;
   int        x=1;
   bool       bBeep;
   
   beep++;
   if(beep==1) return 1;
   if(beep==2){
      statusbar->SetText(bhpos::widget::BHSTATUS_STATUS, "Close Drawer");
      col.set_rgb(65535, 0, 0);
      statusbar->SetColour(bhpos::widget::BHSTATUS_STATUS, Gtk::STATE_INSENSITIVE, col);
      eng->Command(bhmfengine::BHMF_CMD_SYSTEM_BEEP, &x);
   }
   if(beep==3){
      statusbar->SetText(bhpos::widget::BHSTATUS_STATUS, "");
   }
   if(beep > 3){
      config.GetOption(MFENG_CONFIG_CUSTDISP_DEVTYPE, &dCustdisp);
      eng->Command(bhmfengine::BHMF_CMD_SYSTEM_BEEP, &x);
   }
   return 0;
}


int CMfPosEngine::CashDStatus()
{  
   unsigned int ret;
   int          state;

   if(dCashd==POST_CASHD_NONE){
      return -1;     
   }
   else if(dCashd==POST_CASHD_DEVICE){
      if(s0target==POST_DEV_CASHD){
         state=DeviceStatus(0, 0);
#ifdef _DEBUGGERY
            g_print("ttyS0 device status returned: %d", state);
#endif
         if(state < 0){
#ifdef _DEBUGGERY
            g_print("ttyS0 device status error: %d - %s", errno, strerror(errno));
#endif
            return -1;
         }
         else if(state){
            return 1;
         }
         return 0;
      }
      if(s1target==POST_DEV_CASHD){
         state=DeviceStatus(1, 0);
#ifdef _DEBUGGERY
            g_print("ttyS1 device status returned: %d", state);
#endif
         if(state < 0){
#ifdef _DEBUGGERY
            g_print("ttyS1 device status error: %d - %s", errno, strerror(errno));
#endif
            return -1;
         }
         else if(state){
            return 1;
         }
         return 0;
      }
   }
   else if(dCashd==POST_CASHD_PRINTER){
      if(printer){
         ret=printer->GetPrinterStatus(PRINTER_GS_DRAWER_OPEN, BhPos::PRN_STATUSMETHOD_NORMAL);
         if(ret){
            return 1;
         }
         return 0;
      }
   }
   else if(dCashd==POST_CASHD_SMART){  // presumed on RI
      if(s0target==POST_DEV_CASHD){
         state=DeviceStatus(0, 0);
#ifdef _DEBUGGERY
            g_print("ttyS0 device status returned: %d", state);
#endif
         if(state < 0){
#ifdef _DEBUGGERY
            g_print("ttyS0 device status error: %d - %s", errno, strerror(errno));
#endif
            return -1;
         }
         else if(state){
            return 1;
         }
         return 0;
      }
      if(s1target==POST_DEV_CASHD){
         state=DeviceStatus(1, 0);
#ifdef _DEBUGGERY
            g_print("ttyS1 device status returned: %d", state);
#endif
         if(state < 0){
#ifdef _DEBUGGERY
            g_print("ttyS1 device status error: %d - %s", errno, strerror(errno));
#endif
            return -1;
         }
         else if(state){
            return 1;
         }
         return 0;
      }
   }
   
   return -2;
}

int CMfPosEngine::CashDrawer()
{
   bhpos::CashdrawerDeviceType dCashd;
   bhpos::PrinterDeviceType    dPrinter;
   
   config.GetOption(MFENG_CONFIG_CASHDRAWER_DEVTYPE, &dCashd);
   switch(dCashd){
   case bhpos::CDDEV_TYPE_NONE:
      return 1;
   case bhpos::CDDEV_TYPE_PRINTER:
   {
      config.GetOption(MFENG_CONFIG_PRINTER_DEVTYPE, &dPrinter);
      if(dPrinter==bhpos::PRNDEV_TYPE_NONE) return 2;
      bhpos::CPrinter *printer=dynamic_cast<bhpos::CPrinter *>(hardware.GetDevice(bhpos::BHTYPE_PRINTER));
      if(!printer) return 3;
      printer->Print(bhpos::PRNCMD_CASHD, 0, NULL);
      // new printer lib threads, preclude the use of in app timer...
      //sigc::slot<bool> timer_slot = sigc::mem_fun(*this, &CMfPosEngine::CashDrawerTimer);
      //Glib::signal_timeout().connect(timer_slot, 100);
   }
      break;
   case bhpos::CDDEV_TYPE_SERIAL:
      break;
   
   case bhpos::CDDEV_TYPE_SMARTSERIAL:
      break;
   }
   return 0;
}




gint CMfPosEngine::::DrawerStatus(gint fireType)
{
   static int beep=0;
   static bool bWaitTransEnd=false;
   bool bBeep=false;
   int cdstatus;
   
   if(bWaitTransEnd){
      if(post->GetStatus()!=POST_STATUS_IDLE) return 1;
      else{
         if(fireType){
   		   Clear();
   	   }
         else{
            SetStatus(prev_status);
         }
         beep=0;
         bWaitTransEnd=false;
		   return 0;
      }
   }
   cdstatus=post->CashDStatus();

   post->GetOption(BH_POST_OPTION_DEV_CASHD_BEEPONOPEN, &bBeep);
   
   beep++;
	if(beep==1) return 1;   // allow time for drawer to open before testing...
   if(cdstatus < 0){       // cash drawer error
      if(fireType){
         Clear();
      }
      beep=0;
      bWaitTransEnd=false;
		return 0;
   }
   
	if(cdstatus==CASHD_STATUS_OPEN){
	   if(beep==2){
   	   SetStatusBar(2, "");
   	}
   	else if(beep==3){
   	   SetStatusBar(2, _("Close Drawer"));
		}
		else if(beep > 3){
		   beep=0;
		   if(bBeep){
   		   gdk_beep();
   		}
		}
		return 1;
	}
	else if(cdstatus==CASHD_STATUS_CLOSED){
      if(post->GetStatus()!=POST_STATUS_IDLE){
         bWaitTransEnd=true;
         return 1;
      }
	   if(fireType){
   		Clear();
   	}
      else{
         SetStatus(prev_status);
      }
      beep=0;
      bWaitTransEnd=false
		return 0;
	}
	return 1;
}

*/

}
