/*
cprinter.cpp
Copyright (C)2002  Al Kingsley

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Don't write to me at: kingsley@bananahead.com 

   Part of:
   BhPos, Copyright (C) 2001-2005 Al Kingsley
   BhPos comes with ABSOLUTELY NO WARRANTY.
*/
#include "config.h"
#include <cstring>

#include <fcntl.h>
#include <unistd.h>
#include <sys/signal.h>
#include <pthread.h>
#include <math.h>
#include <stdio.h>

#include <libintl.h>
#define _(String) gettext(String)

#include "bhprinter.h"

static volatile int     bReturn=0;
static volatile int     AsbStatus=0;
static volatile int     ThreadWaiting=0;
static volatile bool    bCancellation=false;
static pthread_t        asbthread;
static pthread_mutex_t  printmutex;

namespace bhpos{

void *AsbThreadFunc(void *arg)
//void *SlipWaitThread(void *arg)
{
   CPrinter *prn=(CPrinter *) arg;
   bool bSel=false,bLoad=false,bcdsb=false;
   char buff[8], ibuff[8], ret;
   unsigned int          status, t=0, cdp;
   timespec ts1={0,250000000}, ts2;
   BhPrinterStatusType   type=prn->GetStatusType();
   BhPrinterStatusMethod method=prn->GetStatusMethod();
	prn->GetParam(PRN_PARAM_CASHDRAWERSB, &bcdsb);
   prn->GetParam(PRN_PARAM_CASHDRAWERSBPERIOD, &cdp);
   
   if(prn==NULL){
      pthread_mutex_lock(&printmutex);
      ThreadWaiting=0;
      pthread_mutex_unlock(&printmutex);
      return NULL;
   }
   
   if(method!=PRN_STATUSMETHOD_REALTIME){
      pthread_mutex_lock(&printmutex);
      ThreadWaiting=0;
      pthread_mutex_unlock(&printmutex);
      return NULL;
   }
      
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	
	pthread_mutex_lock(&printmutex);
	AsbStatus=0;
   bCancellation=false;
   pthread_mutex_unlock(&printmutex);
   
   switch(type){
   case PRN_STATUSTYPE_SLIPSELECT:
      buff[0]=0x1B;
      buff[2]=0x63;
      buff[3]=0x34;
      buff[4]=0x3C;
	   // select paper stop on slip
	   prn->Output(4, buff);
   
      // select ABS for paper changes
      buff[0]=0x1D;
      buff[1]=0x61;
      buff[2]=0x2F;
      prn->Output(3, buff);
   
      // wait on above paper change
      prn->Input(4, ibuff, true);
      // select slip printing
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      prn->Output(4, buff);
      break;
   
	case PRN_STATUSTYPE_SLIPLOAD:
      // wait for paper inserted
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      prn->Output(4, buff);
      break;
      
   case PRN_STATUSTYPE_SLIPALL:
      break;
   
   case PRN_STATUSTYPE_SLIPREMOVE:
      break;
   
   case PRN_STATUSTYPE_CHEQUESELECT:
	case PRN_STATUSTYPE_CHEQUELOAD:
      break;
   case PRN_STATUSTYPE_CHEQUEALL:
      // select paper stop on slip
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x34;
      buff[3]=0x3C;
	   prn->Output(4, buff);
   /*
      // cmds to slip
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x31;
      buff[3]=0x04;
      prn->Output(4, buff);
   */
      // output to slip
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      prn->Output(4, buff);
      
   
      // real time slip request
      buff[0]=0x10;
      buff[1]=0x04;
      buff[2]=0x05;
      
      // select ABS for paper changes
      //buff[0]=0x1D;
      //buff[1]=0x61;
      //buff[2]=0x2F;
      prn->Output(3, buff);
      /*
      // wait on above paper change
      prn->Input(4, ibuff, true);
      // select slip printing
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      prn->Output(4, buff);
      
      // wait for paper inserted
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      prn->Output(4, buff);*/
      break;
   case PRN_STATUSTYPE_CHEQUEREMOVE:
      buff[0]=0x1D;
      buff[1]=0x72;
      buff[2]=0x01;
      prn->Output(3, buff);
      break;
   
   case PRN_STATUSTYPE_MICRALL:
	case PRN_STATUSTYPE_MICRSELECT:
      // issue select command
      buff[0]=0x1C;
	   buff[1]=0x61;
	   buff[2]=0x30;
	   buff[3]=0x00;
      prn->Output(4, buff);
      // get real-time status
      buff[0]=0x10;
	   buff[1]=0x04;
	   buff[2]=0x08;
	   buff[3]=0x01;
      prn->Output(4, buff);
      ret=0xFF;
      break;
      
	case PRN_STATUSTYPE_MICRLOAD:
      buff[0]=0x10;
	   buff[1]=0x04;
	   buff[2]=0x08;
	   buff[3]=0x01;
      prn->Output(4, buff);
      ret=0xFF;
      break;
   
   case PRN_STATUSTYPE_MICRREMOVE:
      buff[0]=0x1C;
	   buff[1]=0x61;
	   buff[2]=0x32;
      prn->Output(3, buff);
      buff[0]=0x10;
	   buff[1]=0x04;
	   buff[2]=0x08;
	   buff[3]=0x01;
      prn->Output(4, buff);
      ret=0xFF;
      break;
   
   case PRN_STATUSTYPE_CASHDRAWERCLOSED:
   case PRN_STATUSTYPE_CASHDRAWEROPENCLOSED:
      ret=0x00;
      break;
   }
   
	while(42){
      pthread_mutex_lock(&printmutex);
      if(bCancellation){
         ThreadWaiting=0;
         pthread_mutex_unlock(&printmutex);
         if(type==type==PRN_STATUSTYPE_MICRLOAD || type==PRN_STATUSTYPE_MICRALL || type==PRN_STATUSTYPE_SLIPLOAD){
            // cancel waiting state
            buff[0]=0x10;
            buff[1]=0x05;
            buff[2]=0x03;
            prn->Output(3, buff);
         }
         prn->Signal(PRN_STATUSEVENT_MICRCANCELLED, NULL);
         return NULL;
      }
      pthread_mutex_unlock(&printmutex);
      
      switch(type){
      case PRN_STATUSTYPE_SLIPSELECT:
         // wait for slip selected 
         ibuff[3]=0xFF;
         prn->Input(4, ibuff, true);
         if(ibuff[3] & 0x01){
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_SLIPSELECTED, NULL);
            return NULL;
         }
         break;
      
	   case PRN_STATUSTYPE_SLIPLOAD:
         ibuff[2]=0xFF;
         prn->Input(4, ibuff, true);
         if(bSel){
            // wait for slip loaded
            ibuff[3]=0xFF;
            if(ibuff[3] & 0x02){
               pthread_mutex_lock(&printmutex);
               ThreadWaiting=0;
               pthread_mutex_unlock(&printmutex);
               prn->Signal(PRN_STATUSEVENT_SLIPLOADED, NULL);
               return NULL;
            }   
         }
         else{
            if(ibuff[2] & 0x20){
               if(ibuff[3] & 0x01){
                  pthread_mutex_lock(&printmutex);
                  ThreadWaiting=0;
                  pthread_mutex_unlock(&printmutex);
                  buff[0]=0x10;
                  buff[1]=0x05;
                  buff[2]=0x03;
                  prn->Output(3, buff);
                  prn->Signal(PRN_STATUSEVENT_SLIPCANCELLED, NULL);
                  return NULL;
               }
               else{
                  bSel=true;
               }
            }
         }
         break;
         
      case PRN_STATUSTYPE_SLIPALL:
         ibuff[2]=0xFF;
         prn->Input(4, ibuff, true);
         if(bSel){
            if(bLoad){
               ibuff[3]=0xFF;
               if(ibuff[3] & 0x02){
                  pthread_mutex_lock(&printmutex);
                  ThreadWaiting=0;
                  pthread_mutex_unlock(&printmutex);
                  prn->Signal(PRN_STATUSEVENT_SLIPLOADED, NULL);
                  return NULL;
               }
            }
            else{
               // wait for slip loaded
               ibuff[3]=0xFF;
               if(ibuff[3] & 0x02){
                  bLoad=true;
               }
            }   
         }
         else{
            if(ibuff[2] & 0x20){
               if(ibuff[3] & 0x01){
                  pthread_mutex_lock(&printmutex);
                  ThreadWaiting=0;
                  pthread_mutex_unlock(&printmutex);
                  buff[0]=0x10;
                  buff[1]=0x05;
                  buff[2]=0x03;
                  prn->Output(3, buff);
                  prn->Signal(PRN_STATUSEVENT_SLIPCANCELLED, NULL);
                  return NULL;
               }
               else{
                  bSel=true;
               }
            }
         }
         break;
   
      case PRN_STATUSTYPE_SLIPREMOVE:
         ibuff[3]=0;
         ibuff[4]=0;
         prn->Input(4, ibuff, true);
         if(ibuff[3] & 0x20 && ibuff[4] & 0x01){
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_SLIPREMOVED, NULL);
         }
         break;
         
      case PRN_STATUSTYPE_CHEQUESELECT:
	   case PRN_STATUSTYPE_CHEQUELOAD:
         break;
      
      case PRN_STATUSTYPE_CHEQUEALL:
         if(bSel){
            ibuff[0]=0xFF;
            buff[0]=0x1D;
            buff[1]=0x72;
            buff[2]=0x01;
            prn->Output(3, buff);
            prn->Input(1, ibuff, true);
            // wait for slip loaded (TOF sensor)
            if(ibuff[0] & 0x20){
            }
            else{
               pthread_mutex_lock(&printmutex);
               ThreadWaiting=0;
               pthread_mutex_unlock(&printmutex);
               prn->Signal(PRN_STATUSEVENT_CHEQUELOADED, NULL);
               return NULL;
            }   
         }
         else{
            ibuff[0]=0xFF;
            prn->Input(1, ibuff, true);
            if(ibuff[0] & ~0x04){
               bSel=true;
            }
            else{
               buff[0]=0x10;
               buff[1]=0x04;
               buff[2]=0x05;
               prn->Output(3, buff);
            }
         }
         break;
         
      case PRN_STATUSTYPE_CHEQUEREMOVE:
         ibuff[0]=0xFF;
         prn->Input(1, ibuff, true);
         if(ibuff[0] & 0x20){
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_CHEQUEREMOVED, NULL);
            return NULL;
         }
         else{
            buff[0]=0x1D;
            buff[1]=0x72;
            buff[2]=0x01;
            prn->Output(3, buff);
         }
         break;
      
	   case PRN_STATUSTYPE_MICRSELECT:
         prn->Input(1, &ret, true);
         if(!(ret & 0x02)){
            // slip is selected
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_MICRSELECTED, NULL);
            return NULL;
         }
         buff[0]=0x10;
	      buff[1]=0x04;
	      buff[2]=0x08;
	      buff[3]=0x01;
         prn->Output(4, buff);
         ret=0xFF;
         break;
         
	   case PRN_STATUSTYPE_MICRLOAD:
         prn->Input(1, &ret, true);
         if(bSel){
            // recheck selected
            if(ret & 0x02){
               // read bum data...
               while(1){
                  prn->Input(1, &ret, true);
                  if(ret==0){
                     pthread_mutex_lock(&printmutex);
                     ThreadWaiting=0;
                     pthread_mutex_unlock(&printmutex);
                     prn->Signal(PRN_STATUSEVENT_MICRCANCELLED, NULL);
                     return NULL;
                  }
                  ret=0;
               }
            }
            else{
               pthread_mutex_lock(&printmutex);
               ThreadWaiting=0;
               AsbStatus=PRINTER_STATUS_MICRWAIT;
               pthread_mutex_unlock(&printmutex);
               prn->Signal(PRN_STATUSEVENT_MICRLOADED, NULL);
               return NULL;
            }
         }
         else{
            if(!(ret & 0x03)){
               // slip is loaded
               bSel=true;
            }
         }
         buff[0]=0x10;
	      buff[1]=0x04;
	      buff[2]=0x08;
	      buff[3]=0x01;
         prn->Output(4, buff);
         ret=0xFF;
         break;
         
      case PRN_STATUSTYPE_MICRALL:
         prn->Input(1, &ret, true);
         if(bSel){
            // recheck selected
            if(bLoad){
               if(ret & 0x02){
                  // read bum data...
                  while(1){
                     prn->Input(1, &ret, true);
                     if(ret==0){
                        pthread_mutex_lock(&printmutex);
                        ThreadWaiting=0;
                        AsbStatus=PRINTER_STATUS_MICRWAIT;
                        pthread_mutex_unlock(&printmutex);
                        prn->Signal(PRN_STATUSEVENT_MICRCANCELLED, NULL);
                        return NULL;
                     }
                     ret=0;
                  }
               }
               else{
                  pthread_mutex_lock(&printmutex);
                  ThreadWaiting=0;
                  AsbStatus=PRINTER_STATUS_MICRWAIT;
                  pthread_mutex_unlock(&printmutex);
                  prn->Signal(PRN_STATUSEVENT_MICRLOADED, NULL);
                  return NULL;
               }
            }
            else{
               if(!(ret & 0x03)){
                  bLoad=true;
               }
            }
         }
         else{
            if(!(ret & 0x02)){
               bSel=true;
            }
         }
         buff[0]=0x10;
	      buff[1]=0x04;
	      buff[2]=0x08;
	      buff[3]=0x01;
         prn->Output(4, buff);
         ret=0xFF;
         break;
         
      case PRN_STATUSTYPE_MICRREMOVE:
         prn->Input(1, &ret, true);
         if(ret & 0x02){
            // micr is unselected
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_MICRREMOVED, NULL);
            return NULL;
         }
         buff[0]=0x10;
	      buff[1]=0x04;
	      buff[2]=0x08;
	      buff[3]=0x01;
         prn->Output(4, buff);
         ret=0xFF;
         break;
      
      case PRN_STATUSTYPE_CASHDRAWERCLOSED:
      case PRN_STATUSTYPE_CASHDRAWEROPENCLOSED:
         buff[0]=0x1D;
         buff[1]=0x72;
         buff[2]=2;
         prn->Output(3, buff);
			prn->Input(1, &ret, true);
         //printf("prnstatus:%d\n", ret);
         if(ret & 0x01){
            pthread_mutex_lock(&printmutex);
            ThreadWaiting=0;
            pthread_mutex_unlock(&printmutex);
            prn->Signal(PRN_STATUSEVENT_CASHDRAWERCLOSED, NULL);
            return NULL;
         }
         else{
            if(type==PRN_STATUSTYPE_CASHDRAWEROPENCLOSED){
               if(bcdsb){
                  if(t>=cdp){
                     prn->Signal(PRN_STATUSEVENT_CASHDRAWEROPEN, NULL);
                     t=0;
                  }
                  t+=100;
               }
            }
         }
         pthread_mutex_unlock(&printmutex);
         ret=0;
         break;
      }
      nanosleep(&ts1, &ts2);
   }
   ThreadWaiting=0;
   
   return NULL;
}


CPrinter::CPrinter(BhDevType type) : CBhDevice(type)
{
   //dev=d;
   reccharline=0;
   jrncharline=0;
   slipcharline=0;
   recfont=0;
   jrnfont=0;
   slipfont=0;
   charsize=7;
   max_line=0;
   x_res=180;
   y_res=180;
   bCustomerDisplay=false;
   bAutoCustomerDisplay=false;
   custdisp_type=0;
   custdisp_x=0;
   custdisp_y=0;
   bCashDrawer=false;
   bCashDrawerStatusBack=true;
   cashd_statusperiod=1000;
   
   bAutoThread=true;
	
	linebuffer=NULL;
   custdisp_line1=NULL;
   custdisp_line2=NULL;
   
   bPageMode=false;
   bNVRAM=false;
   nvramsize=0;

   bChequePrint=false;
   bChequeRotate=false;
   bBarcode=false;

   bRealTimeStatusIssued=false;
   bWaitingPaperInsert=false;
   bCreditCut=false;
   bCreditSlip=false;
   bAutoCutter=false;
   
   tbuffsize=0;
   memset(tbuff, 0, sizeof(tbuff));
	
	bAbsWatcher=false;
   statusmethod=PRN_STATUSMETHOD_NORMAL;
   statustype=PRN_STATUSTYPE_NONE;

   memset(&chklayout, 0, sizeof(chklayout));
   
   bMICR=false;
   mircft=E13B;
   
   pthread_mutex_init(&printmutex, NULL);
   
   tprinttype=0;
   num_copies=0;
   num_top=0;
   num_end=0;
   num_cont=0;
   num_sheets=0;
   top_yinc=end_yinc=cont_yinc=0;
   topbuff=contbuff=endbuff=NULL;
   top_len=end_len=cont_len=0;
   tp=cp=ep=NULL;
   
   setlocale(LC_ALL, "");
   bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
   textdomain(GETTEXT_PACKAGE);
}

CPrinter::~CPrinter()
{
   pthread_mutex_destroy(&printmutex);
   if(tp) fclose(tp);
   if(cp) fclose(cp);
   if(ep) fclose(ep);
   if(topbuff) delete [] topbuff;
   if(contbuff) delete [] contbuff;
   if(endbuff) delete [] endbuff;
}

bool CPrinter::PrinterExists()
{
   /*int oldflags;
   unsigned char buff[2]={0x1B, 0x40};
   
   oldflags=fcntl(fd, F_GETFL, 0);
   if(oldflags!=-1){
      oldflags|=O_NONBLOCK;
      if(fcntl(fd, F_SETFL, oldflags)!=-1){
         if(Write(0, 2, buff) < 0){
            return false;
         }
      }
      else{
         return false;
      }
      oldflags&=~O_NONBLOCK;
      fcntl(fd, F_SETFL, oldflags); 
   }
   else{
      return false;
   }*/
   return true;
}

int CPrinter::SetParam(BhPrintParam param, void *v)
{
   switch(param){
   case PRN_PARAM_DEFXRES:
      x_res=*((int *) v);
      break;
   case PRN_PARAM_DEFYRES:
      y_res=*((int *) v);
      break;
   case PRN_PARAM_PAGEMODE:
      bPageMode=*((bool *) v);
      break;
   case PRN_PARAM_CASHDRAWER:
      bCashDrawer=*((bool *) v);
      break;
   case PRN_PARAM_CASHDRAWERSB:
      bCashDrawerStatusBack=*((bool *) v);
      break;
   case PRN_PARAM_CASHDRAWERSBPERIOD:
      cashd_statusperiod=*((int *) v);
      break;
   case PRN_PARAM_CUSTDISP:
      bCustomerDisplay=*((bool *) v);
      break;
   case PRN_PARAM_MICR:
      bMICR=*((bool *) v);
      break;
   case PRN_PARAM_MICRFONTYPE:
      mircft=*((MICRFontType *) v);
      break;
   case PRN_PARAM_RECCHARLINE:
      reccharline=*((int *) v);
		if(reccharline!=0){
			if(linebuffer) delete [] linebuffer;
			linebuffer= new char[reccharline];
			if(!linebuffer) return 1;
			memset(linebuffer, 0x20, reccharline);
		}
      break;
      
   case PRN_PARAM_CHARSET:
      SetCharSet(*((int *) v));
      break;
   
   case PRN_PARAM_RECFONT:
      recfont=*((int *) v);
      break;
   case PRN_PARAM_JRNCHARLINE:
      jrncharline=*((int *) v);
      break;
   case PRN_PARAM_JRNFONT:
      jrnfont=*((int *) v);
      break;
   case PRN_PARAM_SLIPCHARLINE:
      slipcharline=*((int *) v);
      break;
   case PRN_PARAM_SLIPFONT:
      slipfont=*((int *) v);
      break;
   case PRN_PARAM_CUSTDISPTYPE:
      custdisp_type=*((int *) v);
      break;
   case PRN_PARAM_CUSTDISPX:
      custdisp_x=*((int *) v);
		if(custdisp_x!=0){
			if(custdisp_line1) delete [] custdisp_line1;
         if(custdisp_line2) delete [] custdisp_line2;
			custdisp_line1= new char[custdisp_x];
         custdisp_line2= new char[custdisp_x];
			if(!custdisp_line1) return 1;
         if(!custdisp_line2) return 1;
			//memset(custdisp_line1, 0x20, custdisp_x);
         //memset(custdisp_line2, 0x20, custdisp_x);
		}
      break;
   case PRN_PARAM_CUSTDISPY:
      custdisp_y=*((int *) v);
      break;
   case PRN_PARAM_CHEQUE:
      bChequePrint=*((bool *) v);
      break;
   case PRN_PARAM_CHEQUEROTATE:
      bChequeRotate=*((bool *) v);
      break;
   case PRN_PARAM_CHEQUEENDORSE:
      bEndorsement=*((bool *) v);
      break;
   case PRN_PARAM_NVRAM:
      bNVRAM=*((bool *) v);
      break;
   case PRN_PARAM_NVRAMSIZE:
      nvramsize=*((int *) v);
      break;
   case PRN_PARAM_BARCODE:
      bBarcode=*((bool *) v);
      break;
   case PRN_PARAM_AUTOPOLE:
      bAutoCustomerDisplay=*((bool *) v);
      break;
	case PRN_PARAM_AUTOCUTTER:
		bAutoCutter=*((bool *) v);
      break;
	case PRN_PARAM_NVRAM_HDRLOGO:
   case PRN_PARAM_NVRAM_FTRLOGO:
   case PRN_PARAM_NVRAM_SLIPLOGO:
   case PRN_PARAM_NVRAM_HDRMSG:
   case PRN_PARAM_NVRAM_FTRMSG:
   case PRN_PARAM_NVRAM_RECHDRLAYOUT:
   case PRN_PARAM_NVRAM_ITEMLAYOUT:
   case PRN_PARAM_NVRAM_PAYSUMLAYOUT:
   case PRN_PARAM_NVRAM_TRANSSUMLAYOUT:
   case PRN_PARAM_NVRAM_SLIPHDRLAYOUT:
   case PRN_PARAM_NVRAM_SLIPITEMLAYOUT:
   case PRN_PARAM_NVRAM_SLIPPAYSUMLAYOUT:
   case PRN_PARAM_NVRAM_SLIPTRANSSUMLAYOUT:
   case PRN_PARAM_NVRAM_JRNHDR:
   case PRN_PARAM_NVRAM_JRNITEM:
   case PRN_PARAM_NVRAM_JRNFTR:
   case PRN_PARAM_NVRAM_CHPAYTO:
   case PRN_PARAM_NVRAM_MAJAMOUNT:
   case PRN_PARAM_NVRAM_MINAMOUNT:
   case PRN_PARAM_NVRAM_CDIDLEMSG:
   case PRN_PARAM_NVRAM_CDPRESALE:
   case PRN_PARAM_NVRAM_CDPOSTSALE:
   case PRN_PARAM_NVRAM_ROLLCNT:
   case PRN_PARAM_NVRAM_MICRCNT:
		break;
   }
   return 0;
}
int CPrinter::GetParam(BhPrintParam param, void *v)
{
   switch(param){
   case PRN_PARAM_DEFXRES:
      *((int *) v)=x_res;
      break;
   case PRN_PARAM_DEFYRES:
      *((int *) v)=y_res;
      break;
   case PRN_PARAM_PAGEMODE:
      *((bool *) v)=bPageMode;
      break;
   case PRN_PARAM_CASHDRAWER:
      *((bool *) v)=bCashDrawer;
      break;
   case PRN_PARAM_CASHDRAWERSB:
      *((bool *) v)=bCashDrawerStatusBack;
      break;
   case PRN_PARAM_CASHDRAWERSBPERIOD:
      *((int *) v)=cashd_statusperiod;
      break;
   case PRN_PARAM_CUSTDISP:
      *((bool *) v)=bCustomerDisplay;
      break;
   case PRN_PARAM_MICR:
      *((bool *) v)=bMICR;
      break;
   case PRN_PARAM_MICRFONTYPE:
      *((MICRFontType *) v)=mircft;
      break;
   case PRN_PARAM_RECCHARLINE:
      *((int *) v)=reccharline;
      break;
   case PRN_PARAM_RECFONT:
      *((int *) v)=recfont;
      break;
   case PRN_PARAM_JRNCHARLINE:
      *((int *) v)=jrncharline;
      break;
   case PRN_PARAM_JRNFONT:
      *((int *) v)=jrnfont;
      break;
   case PRN_PARAM_SLIPCHARLINE:
      *((int *) v)=slipcharline;
      break;
   case PRN_PARAM_SLIPFONT:
      *((int *) v)=slipfont;
      break;
   case PRN_PARAM_CUSTDISPTYPE:
      *((int *) v)=custdisp_type;
      break;
   case PRN_PARAM_CUSTDISPX:
      *((int *) v)=custdisp_x;
      break;
   case PRN_PARAM_CUSTDISPY:
      *((int *) v)=custdisp_y;
      break;
   case PRN_PARAM_CHEQUE:
      *((bool *) v)=bChequePrint;
      break;
   case PRN_PARAM_CHEQUEROTATE:
      *((bool *) v)=bChequeRotate;
      break;
   case PRN_PARAM_CHEQUEENDORSE:
      *((bool *) v)=bEndorsement;
      break;
   case PRN_PARAM_NVRAM:
      *((bool *) v)=bNVRAM;
      break;
   case PRN_PARAM_NVRAMSIZE:
      *((int *) v)=nvramsize;
      break;
   case PRN_PARAM_BARCODE:
      *((bool *) v)=bBarcode;
      break;
   case PRN_PARAM_AUTOPOLE:
      *((bool *) v)=bAutoCustomerDisplay;
      break;
   }
   return 0;
}

// generally these can be set absolutely, but a default setting 
// takes these parameters: 70% description 10% space 20% price
// with 10% indent for second line quantity - this gives 3 tabs
// the third being the @ symbol for multi quantities
void CPrinter::TabSet(int flags)
{
   int    char_line=reccharline;
   float  tab;
   u_char buff[7]={0x1B, 0x44};
   
   if(flags){
   }
   else{
      // get the char/line per default font
      tab=floor(char_line*0.10);
      buff[2]=(unsigned char) tab;
      tab=floor(char_line*0.70);
      buff[3]=(unsigned char) tab-2; // the @ symbol
      tab=floor(char_line*0.80);
      buff[4]=(unsigned char) tab;
      buff[5]=0x00;
      max_line=(int) floor(char_line*0.70);
      Output(6, buff);
   }
}

int CPrinter::SetCharSet(int n)
{
   char buff[4]={ 0x1B, 0x3D, 0x01 };
     
   if(bCustomerDisplay){
      buff[2]=0x03;
   }
   Output(3, buff);
   
   buff[1]=0x52;
   buff[2]=n;
   Output(3, buff);
   
   buff[1]=0x3D;
   buff[2]=0x01;
   Output(3, buff);
}

int CPrinter::Print(BhPrintType type, int len, void *data)
{
   pthread_mutex_lock(&printmutex);
   
   switch(type){
   case PRNCMD_SETUP:
      Setup((char *) data, len);
      break;
   case PRNCMD_INTIALIZE:
      Initialize((char *) data, len);
      break;
   case PRNCMD_NVBITMAP:
      break;
   case PRNCMD_PREPRINT:
      PrePrint((char *) data, len);
      break;
   case PRNCMD_HEADER:
      PrintHeader((HeaderPrint *) data);
      break;
   case PRNCMD_ITEM:
      PrintItem((CSaleItem *) data);
      break;
   case PRNCMD_TRANSSUMMARY:
      PrintTransSummary((TransSummary *) data);
      break;
	case PRNCMD_PAYITEM:
		PrintPayItem(len, (CPayItem *) data);
		break;
	case PRNCMD_CHANGE:
		PrintChange((double *) data);
		break;
   case PRNCMD_CHARGE:
		PrintCharge((ChargeSummary *) data);
		break;
   case PRNCMD_PAYSUMMARY:
      PrintPaySummary((CTransItems *) data);
      break;
   case PRNCMD_FOOTER:
      PrintFooter((char *) data, len);
      break;
   case PRNCMD_POSTPRINT:
      PostPrint((char *) data, len);
      break;
   case PRNCMD_SLIP_PREPRINT:
      SlipPrePrint((char *) data, len);
      break;
   case PRNCMD_SLIP_HEADER:
      SlipHeader((char *) data, len);
      break;
   case PRNCMD_SLIP_ITEM:
      SlipItem((CSaleItem *) data);
      break;
   case PRNCMD_SLIP_FOOTER:
      SlipFooter((char *) data, len);
      break;
   case PRNCMD_SLIP_POSTPRINT:
      SlipPostPrint((char *) data, len);
      break;
	
	case PRNCMD_SLIP_CANCEL:
		SlipCancel();
      break;
		
   case PRNCMD_LABEL_PREPRINT:
      LabelPrePrint((char *) data, len);
      break;
   case PRNCMD_LABEL_ITEM:
      LabelItem((CSaleItem *) data);
      break;
   case PRNCMD_LABEL_POSTPRINT:
      LabelPostPrint((char *) data, len);
      break;
   case PRNCMD_JRN_PREPRINT:
      JournalPrePrint((char *) data, len);
      break;
   case PRNCMD_JRN_ITEM:
      JournalItem((CSaleItem *) data);
      break;
   case PRNCMD_JRN_POSTPRINT:
      JournalPostPrint((char *) data, len);
      break;
   case PRNCMD_END_PREPRINT:
      EndorsementPrePrint((char *) data, len);
      break;
   case PRNCMD_END_ITEM:
      EndorsementItem((CTransItems *) data);
      break;
   case PRNCMD_END_POSTPRINT:
      EndorsementPostPrint((char *) data, len);
      break; 

   case PRNCMD_CHQ_PREPRINT:
      ChequePrePrint((char *) data, len);
      break;
   case PRNCMD_CHQ_ITEM:
      ChequeItem((ChequePrint *) data);
      break;
   case PRNCMD_CHQ_POSTPRINT:
      ChequePostPrint((char *) data, len);
      break;
   case PRNCMD_CHQ_LOADTIMER:
      StatusThread(PRN_STATUSTYPE_CHEQUEALL);
      break;
   case PRNCMD_CHQ_PRINT:
      ChequePrintItem((ChequePrint *) data);
      break;
   case PRNCMD_CHQ_EJECTTIMER:
      StatusThread(PRN_STATUSTYPE_CHEQUEREMOVE);
      break;
   
   case PRNCMD_MICR_PREREAD:
      if(!bMICR) break;
      MicrPreread((char *) data, len);
      break;
   case PRNCMD_MICR_READ:
      if(!bMICR) break;
      return MicrRead();
      break;
   case PRNCMD_MICR_STATUS:
      if(!bMICR) break;
      return MicrStatus();
      break;
    case PRNCMD_MICR_DATA:
      if(!bMICR) break;
      return MicrData((BhChequeFormattedData *) data);
      break;
   case PRNCMD_MICR_POSTREAD:
      if(!bMICR) break;
      MicrPostread((char *) data, len);
      break;
   case PRNCMD_MICR_CANCEL:
      if(!bMICR) break;
      MicrCancel();
      break;
      
   case PRNCMD_BARCODE:
      break;

   case PRNCMD_CDISPCLEAR:
      if(!bCustomerDisplay) break;
      PrintCustDisp(true, "");
      break;
   case PRNCMD_CDISP:
      if(!bCustomerDisplay) break;
      PrintCustDisp(false, (char *) data);
      break;
   case PRNCMD_CDISPCMD:
      if(!bCustomerDisplay) break;
      PrintCustDispCmd(len);
      break;
   case PRNCMD_CDISPPRESALE:
      if(!bCustomerDisplay) break;
      PrintCustDisp(true,(char *) data);
      break;
   case PRNCMD_CDISPPOSTSALE:
      if(!bCustomerDisplay) break;
      PrintCustDisp(true,(char *) data);
      break;
   case PRNCMD_CDISPIDLE:
      if(!bCustomerDisplay) break;
      PrintCustDisp(true,(char *) data);
      break;
   case PRNCMD_CDISPLINE1:
      if(custdisp_line1){
         strncpy(custdisp_line1, (char *) data, custdisp_x);
         PrintCustDisp(1, len);
      }
      break;
   case PRNCMD_CDISPLINE2:
      if(custdisp_line2){
         strncpy(custdisp_line2, (char *) data, custdisp_x);
         PrintCustDisp(2, len);
      }
      break;
      
      
   case PRNCMD_CASHD:
      FireCashDrawer(len);
      break;
   case PRNCMD_CASHDTIMER:
      StatusThread(PRN_STATUSTYPE_CASHDRAWERCLOSED);
      break;
   
   
	case PRNCMD_CREDIT:
		PrintCredit((CreditPrint *) data);
		break;
   
   case PRNCMD_RAW:
      // type becomes data length
      if(data!=NULL) Output(type, (unsigned char *) data);
      break;
   }
   pthread_mutex_unlock(&printmutex);
   return 0;
}

/////////////////////////////////////////////////////////////////
//
//
//
////////////////////////////////////////////////////////////////

int CPrinter::ReadNVRAM(char key1, char key2, char *data)
{
	PrnNVMsg  msg;
	char     *c=data;
	unsigned char buff[10]={0x1D, 0x28, 0x43, 0x05, 0, 0, 2, 0, key1, key2};

   memset(&msg, 0, sizeof(msg));
	Write(0, 10, buff);

	Read(0, sizeof(msg), &msg);
	if(msg.header!=0x37) return -1;
	if(msg.ident!=0x70) return -2;
	if(msg.status==PRN_NOMOREDATA){
		if((unsigned char) msg.data[0]==0xFF) return -3;
		if((unsigned char) msg.data[0]==0x00) return -4;
		buff[0]=0x06;
		Write(0, 1, buff);
		strcpy(data, msg.data);
		return strlen(msg.data);
	}
	strcpy(data, msg.data);
	c+=81;
	while(1){
		buff[0]=0x06;
		Write(0, 1, buff);
		if(Read(0, 80, c) < 80){
			buff[0]=0x06;
			Write(0, 1, buff);
			return strlen(data);
		}
		c+=81;
	}
	return 0;
}

int CPrinter::WriteNVRam(int amount, char key1, char key2, unsigned char *data)
{
   int   amt=amount+5;
	int   hb=(int) floor((float) (amt/256));
	int   lb=amt-(hb*256);
	unsigned char buff[10]={0x1D, 0x28, 0x43, lb, hb, 0, 1, 0, key1, key2};

	Write(0, 10, buff);
	return Write(0, amount, data);
}

int CPrinter::NVRam(bool bUsed)
{
	int cnt=0;
	unsigned char buff[9]={0x1D, 0x28, 0x43, 0x03, 0x00, 0x00, 0x03, 0x00};

	if(bUsed){
		Write(0, 8, buff);
	}
	else{
		buff[6]=0x04;
		Write(0, 8, buff);
	}
	Read(0, 2, buff);
	if(buff[0]==0x37){
		if(bUsed){
			if(buff[1]!=0x28) return -1;
		}
		else{
			if(buff[1]!=0x29) return -1;
		}
		while(buff[cnt]!=0){
			Read(0, 1, &buff[cnt]);
			cnt++;
			if(cnt > 7) return -2;
		}
		buff[cnt]=0;
		return atoi((const char *) buff);
	}
	
	return -3;
}

int CPrinter::Setup(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}

int CPrinter::Initialize(char *d, int len)
{
   if(d) return Output(len, d);
   
   return -1;
}
int CPrinter::PrePrint(char *d, int len)
{
   if(d) return Output(len, d);  
   return -1;
}

int CPrinter::PrintHeader(HeaderPrint *hdr)
{
   char buff[64]={0x1B, 0x61, 0x01};
   
   // select print mech...
   buff[0]=0x1B;
   buff[1]=0x3D;
   buff[2]=0x01;
   Output(3, buff);
   
   buff[0]=0x1B;
   buff[1]=0x61;
   buff[2]=0x01;
	Output(3, (void *) buff);
	
   if(hdr==NULL) return 1;
   
   if(bNVRAM){
      // print logo
      buff[0]=0x1C;
      buff[1]=0x70;
      buff[2]=0x01;
      buff[3]=0x00;
      Output(4, (void *) buff);
   }
   buff[0]=0x0A;
	buff[1]=0x0A;
   Output(2, (void *) buff); 
	
	// if we have barcode caps, wang the transid out
   if(bBarcode){
      buff[0]=0x1D;
      buff[1]=0x68;
      buff[2]=75;
      Output(3, (void *) buff);
      
      sprintf(&buff[4], "%012d", hdr->transID);
      buff[0]=0x1D;
      buff[1]=0x6B;
      buff[2]=67;
      buff[3]=0x0C;
      Output(16, (void *) buff);
   }
   
   // some problems after barcode print with font... reset it...
   buff[0]=0x1B;
   buff[1]=0x4D;
   buff[2]=0;
   Output(3, (void *) buff);
   
   buff[0]=0x0A;
   Output(1, (void *) buff);
	Output(strlen(hdr->datetime), (void *) hdr->datetime);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 

	buff[0]=0x1B;
	buff[1]=0x61;
	buff[2]=0x00; 
	buff[3]=0x0A; 
   Output(4, (void *) buff);
   
   sprintf(buff, "Transaction : %012d", hdr->transID);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff);

   sprintf(buff, "Operator    : %06d", hdr->UserID);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 

   sprintf(buff, "Station     : %06d", hdr->StationID);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;   // line feed
   buff[1]=0x0A;
   Output(2, (void *) buff); 

   return 0;
}

int CPrinter::PrintItem(CSaleItem *si)
{
   if(si==NULL) return 1;
   
   char          buff[64];
   char          buff2[16];
   unsigned int  x;
   
   if(si->GetType()==BHSI_SITYPE_RETURN){
      strcpy(buff, "(R)");
      Output(strlen(buff), (void *) buff);
   }
   else if(si->GetType()==BHSI_SITYPE_LAYBY){
      strcpy(buff, "(L)");
      Output(strlen(buff), (void *) buff);
   }
   else if(si->GetType()==BHSI_SITYPE_RENTAL){
      strcpy(buff, "(T)");
      Output(strlen(buff), (void *) buff);
   }

   if(si->GetQuantity() > 1){
      sprintf(buff, "%s", si->GetRecText().data());
      if(strlen(buff) > max_line){
         buff[max_line]=0;
      }
		
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // print and line feed
      Output(1, (void *) buff);
      sprintf(buff, "   x %d @ %0.2f", si->GetQuantity(), si->GetPrice());
      sprintf(buff2, "%0.2f", si->GetTotal());
		JustifyLine(0, buff, buff2);
      
      if(si->GetDepositType()==BHSI_DEPOSIT_REDEMPTION){
         strcpy(buff, _("Redemption"));
         if(strlen(buff) > max_line){
            buff[max_line]=0;
         }
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0A;   // print and line feed
         Output(1, (void *) buff);
         sprintf(buff, "   x %d @ %0.2f", si->GetQuantity(), si->GetDeposit());
         sprintf(buff2, "%0.2f", (si->GetQuantity()*si->GetDeposit()));
		   JustifyLine(0, buff, buff2);
      }

      if(bAutoCustomerDisplay){
         int z;
         
         buff[0]=0x1B;
         buff[1]=0x3D;
         buff[2]=0x02;
         Output(3, buff);
         // clear display first
         buff[3]=0x0C;
         Output(4, buff);
         strncpy(buff, si->GetRecText().data(), custdisp_x);
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0A;   // print and line feed
         Output(1, (void *) buff);
         
         // unitprice - qty - subtotal
         sprintf(buff, "%0.2f x %d", si->GetPrice(), si->GetQuantity());
         z=strlen(buff);
         sprintf(buff2, "%0.2f", si->GetTotal());
         x=z+strlen(buff2);
         x-=custdisp_x;
         // pos to diff.
         memset(&buff[z], 0x20, (custdisp_x - x));
         buff[custdisp_x-x]=0;
         strcpy(buff, buff2);
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0D;
         
         // return to receipt selection
         buff[0]=0x1B;
         buff[1]=0x3D;
         buff[2]=0x01;
         Output(3, buff);
      }
   }
   else{
      sprintf(buff, "%s", si->GetRecText().data());
		if(si->GetType()==BHSI_SITYPE_RETURN){
         sprintf(buff2, "-%0.2f", si->GetPrice());
      }
      else{
         sprintf(buff2, "%0.2f", si->GetPrice());
      }
		
      if(strlen(buff) > (reccharline-strlen(buff2) +1)){
         buff[reccharline-strlen(buff2) +1]=0;
      }
		JustifyLine(0, buff, buff2);
      
      if(si->GetDepositType()==BHSI_DEPOSIT_REDEMPTION){
         strcpy(buff, _("Redemption"));
         sprintf(buff2, "%0.2f", si->GetDeposit());
		   JustifyLine(0, buff, buff2);
      }
      
      if(bAutoCustomerDisplay){
         buff[0]=0x1B;
         buff[1]=0x3D;
         buff[2]=0x02;
         Output(3, buff);
         // clear display first
         buff[3]=0x0C;
         Output(4, buff);
         strncpy(buff, si->GetRecText().data(), custdisp_x);
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0A;   // print and line feed
         Output(1, (void *) buff);
         
         // right justify for next line price
         buff[0]=0x1B;
         buff[1]=0x61;
         buff[2]=0x02;
         Output(3, buff);
         
         sprintf(buff, "%0.2f", si->GetTotal());
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0D;
         
         buff[0]=0x1B;
         buff[1]=0x61;
         buff[2]=0x00;
         Output(3, buff);
         
         // return to receipt selection
         buff[0]=0x1B;
         buff[1]=0x3D;
         buff[2]=0x01;
         Output(3, buff);
      }
   }
   
   return 0;
}

int CPrinter::PrintTransSummary(TransSummary *ti)
{
   char buff[64], buff2[64];
   int  x, z;
   
   if(ti==NULL) return 1;
   buff[0]=0x0A;   // print and line feed
	buff[1]=0x0A;
   Output(2, (void *) buff); 

	strcpy(buff, "Num Items :");
   sprintf(buff2, "%d", ti->NumItems);
	JustifyLine(0, buff, buff2);
	   
   strcpy(buff, "Subtotal  :");
   sprintf(buff2, "%0.2f", ti->Subtotal);
   JustifyLine(0, buff, buff2);
   
   strcpy(buff, "Discounts :");
   sprintf(buff2, "%0.2f", ti->Discount);
   JustifyLine(0, buff, buff2);
   
   strcpy(buff, "Tax       :");
   sprintf(buff2, "%0.2f", ti->Tax);
   JustifyLine(0, buff, buff2);
	strcpy(buff2, "________");
	JustifyLine(1, NULL, buff2);
	   
   strcpy(buff, "Total     :");
   sprintf(buff2, "%0.2f", (ti->Subtotal + ti->Tax - ti->Discount));
   JustifyLine(0, buff, buff2);
   
   if(bAutoCustomerDisplay){
      buff[0]=0x1B;
      buff[1]=0x3D;
      buff[2]=0x02;
      Output(3, buff);
      // clear display first
      buff[3]=0x0C;
      Output(4, buff);
      
      strcpy(buff, "Num Items:");
      z=strlen(buff);
      sprintf(buff2, "%d", ti->NumItems);
      x=z+strlen(buff2);
      x-=custdisp_x;
      memset(&buff[z], 0x20, (custdisp_x - x));
      buff[custdisp_x-x]=0;
      strcpy(buff, buff2);
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // print and line feed
      Output(1, (void *) buff);
      
      strcpy(buff, "Total   :");
      z=strlen(buff);
      sprintf(buff2, "%0.2f", ti->Total);
      x=z+strlen(buff2);
      x-=custdisp_x;
      
      memset(&buff[z], 0x20, (custdisp_x - x));
      buff[custdisp_x-x]=0;
      strcpy(buff, buff2);
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0D;
      Output(1, (void *) buff);
         
      buff[0]=0x1B;
      buff[1]=0x3D;
      buff[2]=0x01;
      Output(3, buff);
   }
   
   return 0;
}
int CPrinter::PrintPayItem(int num_pay, CPayItem *pi)
{
	if(pi==NULL) return 1;
   
   char buff[64];
	char buff2[64];
   
   if(num_pay > 1){
		strcpy(buff, "Num Pay Types : ");
      sprintf(buff2, "%d", num_pay);
		JustifyLine(0, buff, buff2);
   }
   if(pi->GetID()==PTYPE_CASH){
		strcpy(buff, _("Cash          :"));
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
   else if(pi->GetID()==PTYPE_CHEQUE){
		strcpy(buff, _("Cheque        :"));
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
   else if(pi->GetID()==PTYPE_CREDIT){
		strcpy(buff, _("Credit        :"));
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_SMARTC){
		strcpy(buff, _("Smart Card    :"));
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_CREDITRETURN){
		strcpy(buff, _("Credit Return :"));
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_CREDITSLIP){
		strcpy(buff, _("Credit Slip   :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_GIFTVOUCHER){
		strcpy(buff, _("Gift Voucher  :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_TPVOUCHER){
		strcpy(buff, _("Ext Voucher   :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_INTCOUPON){
		strcpy(buff, _("Int Coupon    :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_TPCOUPON){
		strcpy(buff, _("Ext Coupon    :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_LUNCHSTAMP){
		strcpy(buff, _("Lunch Stamps  :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_INHOUSECHARGE){
		strcpy(buff, _("IH Charge     :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_STORECARD){
		strcpy(buff, _("Store Card    :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_OTHER){
		strcpy(buff, _("Other         :")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_CREDIT_CB){
		strcpy(buff, _("Credit (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_CHEQUE_CB){
		strcpy(buff, _("Cheque (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_SMARTC_CB){
		strcpy(buff, _("Smart (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_TOKEN_CB){
		strcpy(buff, _("Token (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_VOUCHER_CB){
		strcpy(buff, _("Voucher (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_COUPON_CB){
		strcpy(buff, _("Coupon (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	else if(pi->GetID()==PTYPE_OTHER_CB){
		strcpy(buff, _("Other (Cash Back)")); 
      sprintf(buff2, "%0.2f", pi->GetAmount());
   }
	JustifyLine(0, buff, buff2);
   return 0;
}

int CPrinter::PrintChange(double *d)
{
	char buff[64], buff2[64];
	
	strcpy(buff, "Change        :");
	sprintf(buff2, "%0.2f",  *d);
	JustifyLine(0, buff, buff2);
	return 0;
}

int CPrinter::PrintCharge(ChargeSummary *cs)
{
   char buff[64], buff2[64];
	
	strcpy(buff, "In House Charge Summary");
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
   Output(1, (void *) buff);
   strcpy(buff, "Old Balance:   ");
	sprintf(buff2, "%0.2f",  cs->old_balance);
	JustifyLine(0, buff, buff2);
   strcpy(buff, "New Balance:   ");
	sprintf(buff2, "%0.2f",  cs->new_balance);
	JustifyLine(0, buff, buff2);
   strcpy(buff, "Limit:   ");
	sprintf(buff2, "%0.2f",  cs->limit);
	JustifyLine(0, buff, buff2);
   strcpy(buff, "Available:   ");
	sprintf(buff2, "%0.2f", cs->limit-cs->new_balance);
	JustifyLine(0, buff, buff2);
	return 0;
}

int CPrinter::PrintPaySummary(CTransItems *ti)
{
   if(ti==NULL) return 1;
   
   char buff[64];
   bool bCash=false;
   int  num_pay=ti->NumPayment();
   CPayItem pi;
   
   if(num_pay > 1){
      sprintf(buff, "Num Pay Types : %d", ti->NumPayment());
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // print and line feed
      Output(1, (void *) buff);
   }
   while(num_pay){
      if(ti->GetPayItem(num_pay, pi)){
         num_pay--;
         continue;
      }

      if(pi.GetID()==PTYPE_CASH){
         bCash=true;
         sprintf(buff, "Cash : %0.2f", ti->GetNumber(BHNUM_CASHTENDERED));
         Output(strlen(buff), (void *) buff);
         buff[0]=0x0A;   // print and line feed
         Output(1, (void *) buff);
      }
      else if(pi.GetID()==PTYPE_CHEQUE){
         sprintf(buff, "Cheque : %0.2f", pi.GetAmount());
      }
      else if(pi.GetID()==PTYPE_CREDIT){
         sprintf(buff, "Credit : %0.2f", pi.GetAmount());
         bCreditCut=true;
      }
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // print and line feed
      Output(1, (void *) buff);
   }
   if(bCash){
      sprintf(buff, "Change : %0.2f",  ti->GetNumber(BHNUM_CHANGE));
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;
      Output(1, (void *) buff);
   }
   buff[0]=0x0A;   // print and line feed
   buff[1]=0x0A;
   Output(2, (void *) buff);

   return 0;
}

int CPrinter::PrintFooter(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}

int CPrinter::PostPrint(char *d, int len)
{
   char buff[8]={0x0A,0x0A,0x0A,0x0A,0x0A};
   // should feed, and cut
   if(d){
		Output(len, d);
	}
	else{
		Output(8, buff);
	}
   if(bAutoCutter){
      buff[0]=0x1D;
      buff[1]=0x56;
      buff[3]=0x00;
      if(bCreditCut){
         buff[2]=66;
         Output(4, buff);
         buff[0]=0x0A;
         Output(1, buff);
      }
      else{
         buff[2]=65;
         Output(4, buff);
         buff[0]=0x0A;
         buff[1]=0x0A;
         buff[2]=0x0A;
         buff[3]=0x0A;
         Output(4, buff);
      }
   }
   
   return 0;
}

int CPrinter::PrintZTape(HeaderPrint *hdr, BhServiceTransSum *ts)
{
   int  i;
   char buff[64], buff2[64];
   
   if(hdr==NULL || ts==NULL) return 1;
	
   BhServiceTransSumItem *tsi=ts->item;
   
	buff[0]=0x1B;
	buff[1]=0x61;
	buff[2]=0x01; 
	buff[3]=0x0A; 
   Output(4, (void *) buff);
	
   Output(13, (void *) "Z-Tape Output");
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 
	
	buff[0]=0x1B;
	buff[1]=0x61;
	buff[2]=0x00; 
	buff[3]=0x0A; 
   Output(4, (void *) buff);
   
	Output(strlen(hdr->datetime), (void *) hdr->datetime);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 
   
   sprintf(buff, _("Operator    : %06d"), hdr->UserID);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 

   sprintf(buff, _("Station     : %06d"), hdr->StationID);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 
	buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 
   
   strcpy(buff, _("Opening     :"));
   sprintf(buff2, _("%0.2f"), ts->opening_balance);
   JustifyLine(0, buff, buff2);
   
   strcpy(buff, _("Closing     :"));
   sprintf(buff2, _("%0.2f"), ts->closing_balance);
   JustifyLine(0, buff, buff2);
	
	buff[0]=0x0A;   // line feed
   Output(1, (void *) buff); 
   	
	for(i=0; i!=ts->num_items; i++){
		switch(tsi->type){
		case bhpos::PTYPE_CASH:          strcpy(buff, _("Cash         : ")); break;
		case bhpos::PTYPE_CREDIT:        strcpy(buff, _("Credit:      : ")); break;
		case bhpos::PTYPE_CHEQUE:        strcpy(buff, _("Cheque       : ")); break;
		case bhpos::PTYPE_SMARTC:        strcpy(buff, _("Smart        : ")); break;
	   case bhpos::PTYPE_CREDITRETURN:  strcpy(buff, _("Cr Return    : ")); break;
      case bhpos::PTYPE_CREDITSLIP:    strcpy(buff, _("Credit Slip  : ")); break;   
	   case bhpos::PTYPE_GIFTVOUCHER:   strcpy(buff, _("Gift Voucher : ")); break;
	   case bhpos::PTYPE_TPVOUCHER:     strcpy(buff, _("Ex Voucher   : ")); break;
	   case bhpos::PTYPE_INTCOUPON:     strcpy(buff, _("Int Coupon   : ")); break;
      case bhpos::PTYPE_TPCOUPON:      strcpy(buff, _("Ext Coupon   : ")); break;
	   case bhpos::PTYPE_LUNCHSTAMP:    strcpy(buff, _("Lunch Stamps : ")); break;
      case bhpos::PTYPE_INHOUSECHARGE: strcpy(buff, _("IH Charge    : ")); break;
      case bhpos::PTYPE_STORECARD:     strcpy(buff, _("Store Card   : ")); break;
	   case bhpos::PTYPE_OTHER:         strcpy(buff, _("Other        : ")); break;
	   case bhpos::PTYPE_CREDIT_CB:     strcpy(buff, _("Credit (CB)  : ")); break;
	   case bhpos::PTYPE_CHEQUE_CB:     strcpy(buff, _("Cheque (CB)  : ")); break;
	   case bhpos::PTYPE_SMARTC_CB:     strcpy(buff, _("Smart (CB)   : ")); break;
	   case bhpos::PTYPE_TOKEN_CB:      strcpy(buff, _("Token (CBk)  : ")); break;
	   case bhpos::PTYPE_VOUCHER_CB:    strcpy(buff, _("Voucher (CB) : ")); break;
	   case bhpos::PTYPE_COUPON_CB:     strcpy(buff, _("Coupon (CB)  : ")); break;
	   case bhpos::PTYPE_OTHER_CB:      strcpy(buff, _("Other (CB)   : ")); break;
		}
		sprintf(buff2, "%d", tsi->num);
		strcat(buff, buff2);
		sprintf(buff2, "%0.2f", tsi->amount);
		JustifyLine(0, buff, buff2);
		tsi++;
	}

   sprintf(buff, _("Returns      : "));
	sprintf(buff2, "%d", ts->num_returns);
	strcat(buff, buff2);
   sprintf(buff2, _("%0.2f"), ts->total_returns);
	JustifyLine(0, buff, buff2);
      
	sprintf(buff, _("Paid In      : "));
   sprintf(buff2, _("%d"), ts->num_payin);
   strcat(buff, buff2);
	sprintf(buff2, "%0.2f", ts->total_payin);
   JustifyLine(0, buff, buff2);

   sprintf(buff, _("Paid Out     : "));
   sprintf(buff2, "%d", ts->num_payout);
	strcat(buff, buff2);
	sprintf(buff2, "%0.2f", ts->total_payout);
   JustifyLine(0, buff, buff2);
   
   // 8 lines to cut pos...
	if(bAutoCutter){
      buff[0]=0x1D;
      buff[1]=0x56;
		buff[2]=65;
      buff[3]=0x00;
      Output(4, buff);
      buff[0]=0x0A;
      buff[1]=0x0A;
      buff[2]=0x0A;
      buff[3]=0x0A;
      Output(4, buff);
   }
	/*
   buff[0]=0x0A;
   buff[1]=0x0A;
   buff[2]=0x0A;
   buff[3]=0x0A;
   buff[4]=0x0A;
   buff[5]=0x0A;
   buff[6]=0x0A;
   buff[7]=0x0A;
   Output(8, buff);*/
      
   return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Slip ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::SlipPrePrint(char *d, int len)
{
   unsigned char buff[]={0x1B, 0x2B, 0x41, 0x03};
   
   return 0;
}
int CPrinter::SlipHeader(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}
int CPrinter::SlipItem(CSaleItem *si)
{
   return 0;
}
int CPrinter::SlipFooter(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}
int CPrinter::SlipPostPrint(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}

int CPrinter::SlipCancel()
{
   //pthread_create(&asbthread, NULL, AsbThreadFunc, this);
   //unsigned char buff[3]= {0x10, 0x05, 0x03};
   //Output(3, buff);
   return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Label ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::LabelPrePrint(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}
int CPrinter::LabelItem(CSaleItem *si)
{
   return 0;
}
int CPrinter::LabelPostPrint(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Journal ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::JournalPrePrint(char *d, int len)
{
   return 0;
}
int CPrinter::JournalItem(CSaleItem *si)
{
   return 0;
}
int CPrinter::JournalPostPrint(char *d, int len)
{
   return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Endorsement ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::EndorsementPrePrint(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}
int CPrinter::EndorsementItem(CTransItems *ti)
{
   return 0;
}
int CPrinter::EndorsementPostPrint(char *d, int len)
{
   if(d) return Output(len, d);
   return -1;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Cheque ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int CPrinter::ChequeLayoutToUnit(ChequeLayout *cl, BhChequeLayoutConversion flags)
{
   ChequeUnitLayout cul;
   
   if(cl==NULL) return 1;
   
   if(ChequeLayoutToUnit(cl, &cul, flags)) return 1;
   
   if(ChequeUnitToByte(&cul, &chklayout)) return 2;
   
   return 0;
}

int CPrinter::ChequeLayoutToUnit(ChequeLayout *cl, ChequeUnitLayout *cul, BhChequeLayoutConversion flags)
{
   if(cl==NULL || cul==NULL) return 1;

   float y_resolution;
   float x_resolution;

   if(flags==CHK_LAYOUT_PAGE){
      y_resolution=1.0;
      x_resolution=1.0;
   }
   else if(flags==CHK_LAYOUT_MM){
      y_resolution=25.4/cl->y_dpi;
      x_resolution=25.4/cl->x_dpi;
		printf("xres=%f yres=%f\n", x_resolution, y_resolution);
   }
   else if(flags==CHK_LAYOUT_INCH){
      y_resolution=1.0/cl->y_dpi;
      x_resolution=1.0/cl->x_dpi;
   }
   else{
      return 2;
   }

   cul->size.x=(int) floor((double) cl->size.x/y_resolution);
   cul->size.y=(int) floor((double) cl->size.y/x_resolution);

	//printf("xsize=%d ysize=%d\n", cul->size.x, cul->size.y);
	
   cul->date.x=(int) floor(cl->date.x/y_resolution);
   cul->date.y=(int) floor(cl->date.y/x_resolution);

   cul->year.x=(int) floor(cl->year.x/y_resolution);
   cul->year.y=(int) floor(cl->year.y/x_resolution);

   cul->number_amount.x=(int) floor(cl->number_amount.x/y_resolution);
   cul->number_amount.y=(int) floor(cl->number_amount.y/x_resolution);

   cul->number_word1.x=(int) floor(cl->number_word1.x/y_resolution);
   cul->number_word1.y=(int) floor(cl->number_word1.y/x_resolution);

   cul->number_word1_end.x=(int) floor(cl->number_word1_end.x/y_resolution);
   cul->number_word1_end.y=(int) floor(cl->number_word1_end.y/x_resolution);

   cul->number_word2.x=(int) floor(cl->number_word2.x/y_resolution);
   cul->number_word2.y=(int) floor(cl->number_word2.y/x_resolution);

   cul->payto.x=(int) floor(cl->payto.x/y_resolution);
   cul->payto.y=(int) floor(cl->payto.y/x_resolution);
   
   return 0;
}

int CPrinter::ChequeUnitToByte(ChequeUnitLayout *cul, ChequeByteUnitLayout *cbl)
{
   if(cul==NULL || cbl==NULL) return 1;

	// NB swapped - put back...
   cbl->size.x[0]=cul->size.x%256;
   cbl->size.x[1]=(int) floor((double) cul->size.x/256);
   cbl->size.y[0]=cul->size.y%256;
   cbl->size.y[1]=(int) floor((double) cul->size.y/256);

	//printf("SIZE: x=%d,%d y=%d,%d\n", cbl->size.x[0], cbl->size.x[1], cbl->size.y[0], cbl->size.y[1]);
	
   cbl->date.x[0]=cul->date.x%256;
   cbl->date.x[1]=(int) floor((double) cul->date.x/256);
   cbl->date.y[0]=cul->date.y%256;
   cbl->date.y[1]=(int) floor((double) cul->date.y/256);
	
   //printf("DATE: x=%d,%d y=%d,%d\n", cbl->date.x[0], cbl->date.x[1], cbl->date.y[0], cbl->date.y[1]);
   
	/*
	cbl->size.y[0]=cul->size.x%256;
   cbl->size.y[1]=(int) floor((double) cul->size.x/256);
   cbl->size.x[0]=cul->size.y%256;
   cbl->size.x[1]=(int) floor((double) cul->size.y/256);

	g_print("x=%d,%d y=%d,%d\n", cbl->size.x[0], cbl->size.x[1], cbl->size.y[0], cbl->size.y[1]);
	
   cbl->date.x[0]=cul->date.x%256;
   cbl->date.x[1]=(int) floor((double) cul->date.x/256);
   cbl->date.y[0]=cul->date.y%256;
   cbl->date.y[1]=(int) floor((double) cul->date.y/256);*/

   cbl->year.x[0]=cul->year.x%256;
   cbl->year.x[1]=(int) floor((double) cul->year.x/256);
   cbl->year.y[0]=cul->year.y%256;
   cbl->year.y[1]=(int) floor((double) cul->year.y/256);
	
	//printf("YEAR: yearx=%d,%d yeary=%d,%d\n", cbl->year.x[0], cbl->year.x[1], cbl->year.y[0], cbl->year.y[1]);

   cbl->number_amount.x[0]=cul->number_amount.x%256;
   cbl->number_amount.x[1]=(int) floor((double) cul->number_amount.x/256);
   cbl->number_amount.y[0]=cul->number_amount.y%256;
   cbl->number_amount.y[1]=(int) floor((double) cul->number_amount.y/256);
   
   //printf("AMOUNT: x=%d,%d y=%d,%d\n", cbl->number_amount.x[0], cbl->number_amount.x[1], cbl->number_amount.y[0], cbl->number_amount.y[1]);
   
   cbl->number_word1.x[0]=cul->number_word1.x%256;
   cbl->number_word1.x[1]=(int) floor((double) cul->number_word1.x/256);
   cbl->number_word1.y[0]=cul->number_word1.y%256;
   cbl->number_word1.y[1]=(int) floor((double) cul->number_word1.y/256);

   cbl->num_word1_chars=(int) floor((double) cul->number_word1_end.x/charsize);

   cbl->number_word2.x[0]=cul->number_word2.x%256;
   cbl->number_word2.x[1]=(int)floor((double) cul->number_word2.x/256);
   cbl->number_word2.y[0]=cul->number_word2.y%256;
   cbl->number_word2.y[1]=(int)floor((double) cul->number_word2.y/256);
   
   cbl->payto.x[0]=cul->payto.x%256;
   cbl->payto.x[1]=(int)floor((double) cul->payto.x/256);
   cbl->payto.y[0]=cul->payto.y%256;
   cbl->payto.y[1]=(int)floor((double) cul->payto.y/256);
   
   //printf("PAYTO: x=%d,%d y=%d,%d\n", cbl->payto.x[0], cbl->payto.x[1], cbl->payto.y[0], cbl->payto.y[1]);
   
   return 0;
}

int CPrinter::ChequePrePrint(char *d, int len)
{
   // should select paper insertion, move to bof pos 
   if(d) return Output(len, d);
   return -1;
}

int CPrinter::ChequeItem(ChequePrint *chprn)
{
   int   len1, len2, len3, n;
   char  buff[64]={0x1B, 0x63, 0x33, 0x30};
   char *bptr, *cptr1, *cptr2, *cptr3;
   bool  bl3=false;
   char  line1[128];
   char  line2[128];
   
   bptr=cptr1=cptr2=cptr3=NULL;
	
	Output(4, buff);
   
   if(!bPageMode){   // standard print
      buff[0]=0x1B;      // select slip
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x04;
      Output(4, (void *) buff);
   
      buff[0]=0x1B;   // font select (only works for 9x9 chars)
      buff[1]=0x4D;
      buff[2]=0x00;
      Output(3, (void *) buff);
      
      if(bChequeRotate){
         buff[0]=0x1B;   // line spacing (which will be char spacing) n/180 inch
         buff[1]=0x33;
         buff[2]=0x25;
         Output(3, (void *) buff);
      
         buff[0]=0x1B;   // char spacing (which will be line spacing) n/180 inch
         buff[1]=0x33;
         buff[2]=0x25;
         Output(3, (void *) buff);
      
         buff[0]=0x1B;   // 90 degree rotate
         buff[1]=0x56;
         buff[2]=0x01;
         Output(3, (void *) buff);
      
         //SpellMoney(chprn->amount, line2);
      
         len1=strlen(chprn->payto);
         cptr1=chprn->payto;
      
         if(strlen(line2) > slipcharline){ // line 3 is in play
            cptr2=line2;
            cptr3=&line2[slipcharline];
            bl3=true;
            len2=slipcharline;
            len3=strlen(line2)-slipcharline;
         }
         else{
            cptr2=line2;
            len2=strlen(line2);
            len3=0;
         }
         while(1){
            bptr=buff;
            n=0;
         
            if(len1){
               *bptr=*cptr1;
               bptr++;
               cptr1++;
               len1--;
               n++;
            }
            if(len1){
               *bptr=*cptr2;
               bptr++;
               cptr2++;
               len2--;
               n++;
            }
            if(len3){
               *bptr=*cptr3;
               bptr++;
               cptr3++;
               len1--;
               n++;
            }
            *bptr=0x0A;
            n++;
            Output(n, (void *) buff);
         }
   
         buff[0]=0x1B;   // 90 degree rotate off 
         buff[1]=0x56;
         buff[2]=0x00;
      }
      else{   // forward fed
         //SpellMoney(chprn->amount, line2);
      
         len1=strlen(chprn->payto);
         cptr1=chprn->payto;
      
         if(strlen(line2) > slipcharline){
         }
      }  
      
      buff[0]=0x1B;   // select paper roll
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x01;
      Output(4, (void *) buff);
         
   }
   else{   // page mode... much simpler
		if(!chprn->preLoaded){ // could have come from a preloaded MICR read
			/*
			buff[0]=0x1B;      // enable sensors
	      buff[1]=0x63;
   	   buff[2]=0x34;
      	buff[3]=0x30;
      	Output(4, (void *) buff);*/
			
			buff[0]=0x1B;      // select slip
	      buff[1]=0x63;
   	   buff[2]=0x31;
      	buff[3]=0x04;
      	Output(4, (void *) buff);
			
			buff[0]=0x1B;      // wait for insertion (led blinks)
	      buff[1]=0x63;
   	   buff[2]=0x30;
      	buff[3]=0x04;
      	Output(4, (void *) buff);
		}

      buff[0]=0x1D;
      buff[1]=0x28;
      buff[2]=0x47; 
      buff[3]=0x02; 
      buff[4]=0x00; 
      buff[5]=0x48; 
      buff[6]=0x04; // cut sheet control - select front of paper
      Output(7, buff);
   
      buff[2]=0x02;
      buff[3]=0x00;
      buff[4]=0x84;
      buff[5]=0x01;
      Output(6, buff);  // cut sheet control - move to print start pos

      buff[0]=0x1B;
      buff[1]=0x4C;  // page mode
      Output(2, buff);

      // set print area
      buff[1]=0x57;
      buff[2]=0x00;
      buff[3]=0x00;
      buff[4]=0x00;
      buff[5]=0x00;
      memcpy(&buff[6], (unsigned char *) &chklayout.size, 4);
      Output(10, buff);

      // move to date position
      // vertically in print (horz for real)
      // rotate 90 degrees
      buff[0]=0x1B;
      buff[1]=0x54;
      buff[2]=0x01;
      Output(3, buff);

      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.date.y, 2);
      Output(4, buff);
      		
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.date.x, 2);
      Output(4, buff);
		//g_print("nobber 1-%d|%d,%d|%d\n", chklayout.date.x[0], chklayout.date.x[1], chklayout.date.y[0], chklayout.date.y[1]);
		//Output(7, (void *) "fuckpit");
      Output(strlen(chprn->date), chprn->date);

      // move to year 
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.year.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.year.x, 2);
      Output(4, buff);
      Output(strlen(chprn->year), chprn->year);
		
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.payto.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.payto.x, 2);
      Output(4, buff);
      Output(strlen(chprn->payto), chprn->payto);
      
		//buff[5]=0;
		//g_print("nobber 1-%d|%d,%d|%d - %s\n", chklayout.year.x[0], chklayout.year.x[1], chklayout.year.y[0], chklayout.year.y[1], buff);

      // move to num amount
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_amount.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_amount.x, 2);
      Output(4, buff);
      sprintf(buff, "%0.2f", chprn->amount);
      Output(strlen(buff), buff);

      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_word1.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_word1.x, 2);
      Output(4, buff);

      SpellMoney(chprn->amount, line1, chprn->MajAmount, chprn->MinAmount);
      // check if double liner...
      len1=strlen(line1);
      if(len1 > chklayout.num_word1_chars){
         cptr1=&line1[chklayout.num_word1_chars];
         while(*cptr1!=' '){
            cptr1--;
         }
         strcpy(line2, cptr1+1);
         *cptr1=0;
         Output(strlen(line1), line1);

         buff[0]=0x1D;
         buff[1]=0x24;
         memcpy(&buff[2], chklayout.number_word2.y, 2);
         Output(4, buff);
         buff[0]=0x1B;
         buff[1]=0x24;
         memcpy(&buff[2], chklayout.number_word2.x, 2);
         Output(4, buff);
         Output(strlen(line2), line2);
      }
      else{
         Output(strlen(line1), line1);
      }
 
      

      // print it and return to normal mode, 
      // select receipt paper.
      // eject

      buff[0]=0x0C;
      Output(1, buff);


      //buff[0]=0x1B;
      //buff[1]=0x53;
      //Output(2, buff);
   }

   return 0;
}
int CPrinter::ChequePostPrint(char *d, int len)
{
   /* eject the cheque
   unsigned char buff[3]= {0x1C, 0x61, 0x32 };
   Output(3, buff);
   */
   
   if(d) return Output(len, d);
   return -1;
}

int CPrinter::ChequePrintItem(ChequePrint *chprn)
{
   int   len1, len2, len3, n;
   char  buff[64]={0x1B, 0x63, 0x33, 0x30};
   char *bptr, *cptr1, *cptr2, *cptr3;
   bool  bl3=false;
   char  line1[128];
   char  line2[128];
   
   bptr=cptr1=cptr2=cptr3=NULL;
	
   if(!bPageMode){   // standard print
      printf("yo ho ho \n");
      buff[0]=0x1B;   // font select (only works for 9x9 chars)
      buff[1]=0x4D;
      buff[2]=0x00;
      Output(3, (void *) buff);
      bChequeRotate=true;
      if(bChequeRotate){
         printf("ho ho ho \n");
         buff[0]=0x1B;   // line spacing (which will be char spacing) n/180 inch
         buff[1]=0x33;
         buff[2]=0x25;
         Output(3, (void *) buff);
      
         buff[0]=0x1B;   // char spacing (which will be line spacing) n/180 inch
         buff[1]=0x33;
         buff[2]=0x25;
         Output(3, (void *) buff);
      
         buff[0]=0x1B;   // 90 degree rotate
         buff[1]=0x56;
         buff[2]=0x01;
         Output(3, (void *) buff);
      
         //SpellMoney(chprn->amount, line2);
      
         len1=strlen(chprn->payto);
         cptr1=chprn->payto;
      
         if(strlen(line2) > slipcharline){ // line 3 is in play
            cptr2=line2;
            cptr3=&line2[slipcharline];
            bl3=true;
            len2=slipcharline;
            len3=strlen(line2)-slipcharline;
         }
         else{
            cptr2=line2;
            len2=strlen(line2);
            len3=0;
         }
         while(1){
            bptr=buff;
            n=0;
         
            if(len1){
               *bptr=*cptr1;
               bptr++;
               cptr1++;
               len1--;
               n++;
            }
            if(len1){
               *bptr=*cptr2;
               bptr++;
               cptr2++;
               len2--;
               n++;
            }
            if(len3){
               *bptr=*cptr3;
               bptr++;
               cptr3++;
               len1--;
               n++;
            }
            *bptr=0x0A;
            n++;
            Output(n, (void *) buff);
         }
   
         buff[0]=0x1B;   // 90 degree rotate off 
         buff[1]=0x56;
         buff[2]=0x00;
      }
      else{   // forward fed
         //SpellMoney(chprn->amount, line2);
      
         len1=strlen(chprn->payto);
         cptr1=chprn->payto;
      
         if(strlen(line2) > slipcharline){
         }
         //Output(strnlen(line1), line1);
      }
      buff[0]=0x1B;
      buff[1]=0x63;
      buff[2]=0x30;
      buff[3]=0x01;
      Output(4, buff);
   }
   else{   // page mode... much simpler
		if(!chprn->preLoaded){ // could have come from a preloaded MICR read
		}
      
      printf("3\n");
      /*
      buff[0]=0x1D;
      buff[1]=0x28;
      buff[2]=0x47; 
      buff[3]=0x02; 
      buff[4]=0x00; 
      buff[5]=0x48; 
      buff[6]=0x04; // cut sheet control - select front of paper
      Output(7, buff);
   
      buff[2]=0x02;
      buff[3]=0x00;
      buff[4]=0x84;
      buff[5]=0x01;
      Output(6, buff);  // cut sheet control - move to print start pos
      */
      buff[0]=0x1B;
      buff[1]=0x4C;  // page mode
      Output(2, buff);

      // set print area
      buff[1]=0x57;
      buff[2]=0x00;
      buff[3]=0x00;
      buff[4]=0x00;
      buff[5]=0x00;
      memcpy(&buff[6], (unsigned char *) &chklayout.size, 4);
      Output(10, buff);

      // move to date position
      // vertically in print (horz for real)
      // rotate 90 degrees
      buff[0]=0x1B;
      buff[1]=0x54;
      buff[2]=0x01;
      Output(3, buff);

      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.date.y, 2);
      Output(4, buff);
      		
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.date.x, 2);
      Output(4, buff);
		//printf("DATE %d|%d,%d|%d\n", chklayout.date.x[0], chklayout.date.x[1], chklayout.date.y[0], chklayout.date.y[1]);
		//Output(7, (void *) "fuckpit");
      Output(strlen(chprn->date), chprn->date);

      // move to year 
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.year.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.year.x, 2);
      Output(4, buff);
      Output(strlen(chprn->year), chprn->year);

      // move to num amount
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_amount.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_amount.x, 2);
      Output(4, buff);
      sprintf(buff, "%0.2f", chprn->amount);
      Output(strlen(buff), buff);

      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_word1.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.number_word1.x, 2);
      Output(4, buff);

      SpellMoney(chprn->amount, line1, chprn->MajAmount, chprn->MinAmount);
      // check if double liner...
      len1=strlen(line1);
      if(len1 > chklayout.num_word1_chars){
         cptr1=&line1[chklayout.num_word1_chars];
         while(*cptr1!=' '){
            cptr1--;
         }
         strcpy(line2, cptr1+1);
         *cptr1=0;
         Output(strlen(line1), line1);

         buff[0]=0x1D;
         buff[1]=0x24;
         memcpy(&buff[2], chklayout.number_word2.y, 2);
         Output(4, buff);
         buff[0]=0x1B;
         buff[1]=0x24;
         memcpy(&buff[2], chklayout.number_word2.x, 2);
         Output(4, buff);
         Output(strlen(line2), line2);
      }
      else{
         Output(strlen(line1), line1);
      }
 
      buff[0]=0x1D;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.payto.y, 2);
      Output(4, buff);
      buff[0]=0x1B;
      buff[1]=0x24;
      memcpy(&buff[2], chklayout.payto.x, 2);
      Output(4, buff);
      Output(strlen(chprn->payto), chprn->payto);

      // print it and return to normal mode, 
      // select receipt paper.
      // eject

      buff[0]=0x0C;
      Output(1, buff);
      
      buff[0]=0xFF;
      Output(1, buff);


      //buff[0]=0x1B;
      //buff[1]=0x53;
      //Output(2, buff);
   }

   return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Credit ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::PrintCredit(CreditPrint *cr)
{
	char buff[64]={0x1B, 0x61, 0x01};
	
	Output(3, buff);
	Output(strlen(cr->header), (void *) cr->header);
	buff[0]=0x0A;
   Output(1, buff);
	
	Output(strlen(cr->datetime), (void *) cr->datetime);
   buff[0]=0x0A;
   Output(1, buff); 
   
	buff[0]=0x1B;
	buff[1]=0x61;
	buff[2]=0x00; 
	buff[3]=0x0A; 
   Output(4, (void *) buff);
   
   sprintf(buff, "Transaction  : %012d", cr->trans);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
   Output(1, (void *) buff);

   sprintf(buff, "Operator     : %06d", cr->user);
   Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
   Output(1, (void *) buff); 

   sprintf(buff, "Station      : %06d", cr->station);
   Output(strlen(buff), (void *) buff);
   buff[1]=0x0A;
   Output(2, (void *) buff);
	
	sprintf(buff, "Authorisation : %016d", cr->auth);
	Output(strlen(buff), (void *) buff);
   buff[0]=0x0A; 
   Output(1, (void *) buff);
	
	sprintf(buff, "Card Type     : %s", cr->cardtype);
	Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
   Output(1, (void *) buff);
	
	sprintf(buff, "Card Number   : %s", cr->cardnum);
	Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
   Output(1, (void *) buff);
   
	sprintf(buff, "Amount       : %0.2f", cr->amount);
	Output(strlen(buff), (void *) buff);
   buff[0]=0x0A;
	buff[1]=0x0A;
   Output(2, (void *) buff);
	
	if(cr->bSigline){
		buff[0]=0x0A;
		buff[1]=0x0A;
   	Output(2, (void *) buff);
		memset(buff, 95, reccharline);
		memcpy(buff, "Signature", 9);
		buff[0]=0x0A;
   	Output(1, (void *) buff);
	}
   buff[0]=0x1D;
   buff[1]=0x56;
   buff[2]=0x66;
   buff[3]=0x00;
   Output(4, buff);
   
   // 14mm feed - or 4 lines - to manual cut postion
   buff[0]=0x0A;
   buff[1]=0x0A;
   buff[2]=0x0A;
   buff[3]=0x0A;
   Output(4, buff);
	
   bCreditCut=false;
   
	return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ MICR ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::MicrPreread(char *d, int len)
{
   // monitor TOF / BOF / select MICR
   unsigned char buff[10]= {0x1C, 0x61, 0x30, 0x00};
	unsigned char ibuff[4];
   
   Output(4, buff);
	
   // wait on MICR selected 
	buff[0]=0x10;
	buff[1]=0x04;
	buff[2]=0x08;
	buff[3]=0x01;
   Output(4, buff);
	
   ibuff[0]=0xFF;
   while(ibuff[0] & 0x02){
      Input(1, ibuff, true);
      sleep(100);
      Output(4, buff);
   }
   // wait on check loaded & check cancellation
   ibuff[0]=0xFF;
   while(ibuff[0] & 0x04){
      Input(1, ibuff, true);
      if(ibuff[0] & 0x02){
         MicrCancel();
         return 0;
      }
      sleep(100);
      Output(4, buff);
   }

   return 0;
}

int CPrinter::MicrCancel()
{
   unsigned char buff[3]= {0x10, 0x05, 0x03};
   bReturn=1;
   Output(3, buff);
   return 0;
}

int CPrinter::MicrRead()
{
   unsigned char buff[64]= {0x1C, 0x61, 0x30, 0x00};
   char          status;
   int           i=0;
   
   // select cheque read
   //buff[3]=(unsigned char) mircft;
   Output(4, buff);
   Input(1, buff, true);
   if(buff[0]!=0x5F) return 1;
   Input(1, buff, true);
   status=buff[0];
   
   if(status & 0x20) return 2;  // bum read
   while(1){
      Input(1, buff, true);
      if(buff[i]==0){
         break;
      }
      i++;
      buff[i]=0;
   }
   memcpy(tbuff, buff, i);
   tbuffsize=i;
   
   return 0;
}

unsigned char CPrinter::MicrStatus()
{
   unsigned char buff[4]= { 0x10, 0x04, 0x08, 0x01 };

   Output(4, buff);
   Input(1, buff, false);

   return buff[0];
}

int CPrinter::MicrData(BhChequeFormattedData *chkdata)
{
   char *c=tbuff;
   char  buff[32];
   int   i, cnt=0;
   bool  bdig1=true, bTransit1=false, bAuxOnUs=false, bOnUs=false, bAmount=false;

   /*if(chkdata->header!=0x5F) return 1;

   if(chkdata->status & 0x20){ // bad read
      return 2;
   }*/
   
   while(c){
      if(bdig1){
         if(*c=='T'){ // personal cheque transit
            bdig1=false;
            bTransit1=true;
            c++;
            continue;
         }
         else if(*c=='O'){ // corporate aux on us
            bdig1=false;
            bAuxOnUs=true;
            c++;
            continue;
         }
         else return 3;
      }

      if(bTransit1){
         if(cnt==9 && *c=='T'){   // done
            buff[cnt]=0;
            strcpy(chkdata->transit.transit_r, buff);
            cnt=0;
            bTransit1=false;
            c++;
            continue;
         }
         if(cnt>=9) return 4;
         if(*c < '0' || *c >'9') return 5;
         buff[cnt]=*c;
         cnt++;
         c++;
         continue;
      }
      if(bAuxOnUs){
         if(cnt>=11) return 6;
         if(*c=='O'){ // done
            buff[cnt]=0;
            strcpy(chkdata->auxonus, buff);
            cnt=0;
            bAuxOnUs=false;
            c++;
            continue;
         }
         if(*c < '0' || *c >'9') return 7;
         buff[cnt]=*c;
         cnt++;
         c++;
         continue;
      }

      if(bOnUs){  // variable length field - buffer until Amount field
         if(cnt>=19) return 8;
         if(*c=='A'){ // we have an account field - back track
            buff[cnt]=0;
            strcpy(chkdata->onus, buff);
            cnt=0;
            bOnUs=false;
            continue;
         }
         if(*c < '0' || *c >'9' || *c!=' ' || *c!='D' || *c!='O') return 9;
         buff[cnt]=*c;
         cnt++;
         c++;
         continue;
      }

      if(bAmount){
         if(cnt>=10) return 10;
         if(*c=='A'){ // done
            buff[cnt]=0;
            strcpy(chkdata->amount, buff);
            cnt=0;
            bOnUs=false;
            c++;
            continue;
         }
         if(*c < '0' || *c >'9') return 11;
         buff[cnt]=*c;
         cnt++;
         c++;
         continue;
      }

      if(*c=='O'){ // standard on us
         bOnUs=true;
         cnt++;
         c++;
         continue;
      }
      if(*c=='A'){ // standard amount
         bAmount=true;
         cnt++;
         c++;
         continue;
      }
      c++;
   }
   
   return 0;
}

int CPrinter::MicrPostread(char *d, int len)
{
   if(tbuffsize!=0){
      memcpy(d, tbuff, tbuffsize);
      memset(tbuff, 0, sizeof(tbuff));
      tbuffsize=0;
      return 0;
   }
   
   // load cheque to print start position...
   unsigned char buff[8]= {0x1C, 0x61, 0x31 };
   
   Output(3, buff);
   
   return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Misc ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int CPrinter::PrintCustDisp(unsigned char disp, int x, int y, CSaleItem *si)
{
   if(custdisp_type==1) return 1;
   
   char buff[64]={ 0x1B, 0x3D, 0x02 };
   
   if(si==NULL) return 1;
   
   // select customer display only
   Output(3, buff);
      
   // clear display first
   buff[3]=0x0C;
   Output(4, buff);
      
   // put some crap up...    
   if(si->GetQuantity() > 1){
      sprintf(buff, "%s", si->GetRecText().data());
      if(strlen(buff) > custdisp_x){
         buff[custdisp_x]=0;
      }
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // print and line feed
      Output(1, (void *) buff);
      
      sprintf(buff, "%0.2f", si->GetPrice());
      Output(strlen(buff), (void *) buff);
      buff[0]=0x0A;   // line feed
      Output(1, (void *) buff);   // tab
      
      sprintf(buff, "x %d", si->GetQuantity());
      Output(strlen(buff), (void *) buff);
      buff[0]=0x09;
      Output(1, (void *) buff);   // tab
      sprintf(buff, "%0.2f", si->GetTotal());
      Output(strlen(buff), (void *) buff);
   }
   else{
      sprintf(buff, "%s", si->GetRecText().data());
      if(strlen(buff) > custdisp_x){
         buff[custdisp_x]=0;
      }
      Output(strlen(buff), (void *) buff);
      sprintf(buff, "%0.2f", si->GetPrice());
      Output(strlen(buff), (void *) buff);
   }
   buff[0]=0x0A;   // print and line feed
   Output(1, (void *) buff);
   
   // select the printer device only (standard mode)
   buff[0]=0x1B;
   buff[1]=0x3D;
   buff[2]=0x01;
   Output(3, buff);
   return 0;
}

int CPrinter::PrintCustDisp(bool b, char *msg)
{
   char buff[64]={ 0x1B, 0x3D, 0x02 };
   
   if(msg==NULL) return 1;
   
   // select customer display only
   Output(3, buff);
   
   // clear display if wanted
   if(b){
      buff[3]=0x0C;
      Output(4, buff);
      // move cursor home
      //buff[0]=0x0B;
      //Output(1, buff);
      //buff[0]=0x1B;
   }
   // dump message
   Output(strlen(msg), (void *) msg);
   
   // select print mech
   buff[2]=0x01;
   Output(3, buff);
   
   return 0;
}

int CPrinter::PrintCustDisp(int l, int b)
{
   char buff[64]={ 0x1B, 0x3D, 0x02 };
   
   // select customer display only
   Output(3, buff);
   
   // clear display if wanted
   if(b!=0){
      buff[3]=0x0C;
      Output(4, buff);
   }
   // dump message
   if(l==1){
      buff[0]=0x1F;
      buff[1]=0x24;
      buff[2]=0x01;
      buff[3]=0x01;
      Output(4, buff);
      Output(strlen(custdisp_line1), (void *) custdisp_line1);
   }
   if(l==2){
      buff[0]=0x1F;
      buff[1]=0x24;
      buff[2]=0x01;
      buff[3]=0x02;
      Output(4, buff);
      Output(strlen(custdisp_line2), (void *) custdisp_line2);
   }
   // select print mech
   buff[0]=0x1B;
   buff[1]=0x3D;
   buff[2]=0x01;
   Output(3, buff);
   
   return 0;
}

int CPrinter::PrintCustDispCmd(int c)
{
   char buff[64]={ 0x1B, 0x3D, 0x02 };
   
   // select customer display only
   Output(3, buff);
   
   // clear display if wanted
   buff[3]=0x0C;
   Output(4, buff);
   
   if(c==1){ // continous time
      buff[0]=0x1F;
      buff[1]=0x55;
      Output(2, buff);
   }
   if(c==2){ // cursor off
      buff[0]=0x1F;
      buff[1]=0x43;
      buff[2]=0x00;
      Output(3, buff);
   }
   
   // select print mech
   buff[0]=0x1B;
   buff[1]=0x3D;
   buff[2]=0x01;
   Output(3, buff);
   
   return 0;
}

int CPrinter::PrintCustDisp(char *msg1, char *msg2)
{
   char buff[4]={ 0x1B, 0x3D, 0x02 };
   char lf=0x0A;
   
   if(msg1==NULL) return 1;
   
   // select customer display only
   Output(3, buff);
   
   // clear display
   buff[3]=0x0C;
   Output(4, buff);

   // dump message
   Output(strlen(msg1), (void *) msg1);
   Output(1, (void *) &lf);
   Output(strlen(msg2), (void *) msg2);
   // select print mech
   buff[2]=0x01;
   Output(3, buff);
   
   return 0;
}

int CPrinter::FireCashDrawer(int type)
{
   unsigned char buff[5]={0x1B, 0x70, 0, 125, 250};
   int ret=Output(5, buff);
   if(type==0){
      return ret;
   }
   else if(type==1){
      StatusThread(PRN_STATUSTYPE_CASHDRAWERCLOSED);
   }
   else if(type==2){
      StatusThread(PRN_STATUSTYPE_CASHDRAWEROPENCLOSED);
   }
   return ret;
   //buff[2]=0x01;
   //return Output(5, buff);
}

int CPrinter::PaperCut(unsigned char type, unsigned char feed)
{
	char buff[4];
	
	buff[0]=0x1D;
   buff[1]=0x56;
	buff[2]=type;
   buff[3]=feed;
   Output(4, buff);
   buff[0]=0x0A;
   buff[1]=0x0A;
   buff[2]=0x0A;
   buff[3]=0x0A;
   Output(4, buff);
}

int CPrinter::LoadSlip()
{
   char  buff[8]={0x1B, 0x63, 0x34, 0x3C};
	char  ibuff[8];
   
   // select paper stop on slip
	Output(4, buff);
   
   // select ABS for paper changes
   buff[0]=0x1D;
   buff[1]=0x61;
   buff[2]=0x2F;
   Output(3, buff);
   
   // wait on above paper change
   Input(4, ibuff, true);
   
   // select slip printing
   buff[0]=0x1B;
   buff[1]=0x63;
   buff[2]=0x30;
   buff[3]=0x04;
   Output(4, buff);
   
   // wait for slip selected 
   ibuff[3]=0xFF;
   while(ibuff[3] & 0x01){
      Input(4, ibuff, true);
      sleep(100);
   }
   
   // wait for paper inserted
   ibuff[2]=0xFF;
   while(ibuff[2] & 0x20){
      Input(4, ibuff, true);
      // check for cancel
      if(ibuff[3] & 0x01){
         CancelSlip();
      }
      sleep(100);
   }
   
   // wait for slip loaded
   ibuff[3]=0xFF;
   while(ibuff[3] & 0x02){
      Input(4, ibuff, true);
      sleep(100);
   }
}

bool CPrinter::CheckSlip()
{
   char  ibuff[8]={0xFF, 0xFF, 0xFF, 0xFF};
   
    // check printing is possible...
   Input(4, ibuff, true);
   if(ibuff[3] & 0x02){
      return true;
   }
   return false;
}

void CPrinter::CancelSlip()
{
   unsigned char buff[3]= {0x10, 0x05, 0x03};

   Output(3, buff);
}

void CPrinter::EjectSlip()
{
   char  buff[8]={0xFF};
	char  ibuff[8]={0xFF, 0xFF, 0xFF, 0xFF};
   
   // eject
	Output(1, buff);
   
   // wait for slip ejected and unselected
   while(!(ibuff[2] & 0x40) && !(ibuff[3] & 0x01)){
      Input(4, ibuff, true);
      sleep(100);
   }
}

/////////////////////////////////////////////////////////////////
//
//
//
////////////////////////////////////////////////////////////////

int CPrinter::Output(int len, void *d)
{ 
   return Write(0, len, d);
}

int CPrinter::Input(int len, void *d, bool bCanBlock)
{
   /*
   int   res;
   fd_set fds;
   struct timeval timeout;
   
   if(bCanBlock){
      return Read(0, len, d);
   }
   
   FD_ZERO(&fds);
   FD_SET(dev, &fds);
   timeout.tv_sec=0;
   timeout.tv_usec=10000;
   
   res=select(dev+1, &fds, NULL, NULL, &timeout);
   if(res==0){ // timeout
      return 0;
   } 
   else if(res < 0){  // Error Occurred
      return -1;
   }
   */
   return Read(0, len, d);
}

void CPrinter::Lock()
{
   pthread_mutex_lock(&printmutex);
}

void CPrinter::Release()
{
   pthread_mutex_unlock(&printmutex);
}

void CPrinter::StatusThread(BhPrinterStatusType type)
{
   statustype=type;
   statusmethod=PRN_STATUSMETHOD_REALTIME;
   pthread_create(&asbthread, NULL, AsbThreadFunc, this);
}

/////////////////////////////////////////////////////////////////
//
//
//
////////////////////////////////////////////////////////////////

unsigned int CPrinter::GetPrinterStatus(unsigned int type, BhPrinterStatusMethod method)
{
   unsigned char buff[10]={0x10, 0x04, 0x01, 0x01};
   unsigned char ret=0;
   unsigned char rat;
   unsigned int  status=0;
	
   return 0;
   statusmethod=method;
   //statustype=type;

   if(method==PRN_STATUSMETHOD_NORMAL){
		buff[0]=0x1D;
      buff[1]=0x72;
		buff[2]=1;
      switch(type){
      case PRINTER_GS_PAPERROLL_NEAREND:
         Output(3, buff);
			Input(1, &ret, true);
         if(ret && 0x03) status=1;
         break;
      case PRINTER_GS_PAPERROLL_PAPER:
         Output(3, buff);
			Input(1, &ret, true);
         if(ret && 0x12) status=1;
         break;
      case PRINTER_GS_PAPERROLL_TOF:
         Output(3, buff);
			Input(1, &ret, true);
         if(ret && 0x32) status=1;
         break;
      case PRINTER_GS_PAPERROLL_BOF:
         Output(3, buff);
			Input(1, &ret, true);
         if(ret && 0x64) status=1;
         break;
      case PRINTER_GS_DRAWER_OPEN:
         buff[2]=2;
         Output(3, buff);
			Input(1, &ret, true);
         if(ret && 0x01) status=1;
         break;
      case PRINTER_GS_SLIPPAPER_PRESENT:
         buff[2]=3;
         Output(3, buff);
			Input(1, &ret, true);
         if(ret!=0x00) status=1;
         break;
      }
      return status;
	}
	
	if(method==PRN_STATUSMETHOD_ASB){
		if(bAbsWatcher){
			pthread_mutex_lock(&printmutex);
			status=AsbStatus;
			pthread_mutex_unlock(&printmutex);
		}
		else{
			if(type & PRINTER_GETSTATUS_SLIP){
				bAbsWatcher=true;
				/*
				buff[0]=0x1D;
				buff[1]=0x28;
				buff[2]=0x40;
				buff[3]=5;
				buff[4]=0;
				buff[5]=97;
				buff[6]=49;
				buff[7]=44;
				buff[8]=51;
				buff[9]=45;
				Output(10, buff);*/
				//pthread_mutex_lock(&printmutex);
				pthread_create(&asbthread, NULL, AsbThreadFunc, this);
				//pthread_mutex_unlock(&printmutex);
				sleep(1);
				buff[0]=0x1D;
				buff[1]=0x61;
				buff[2]=0x20;
				Output(3, buff);
				
				//Input(4, buff, true);
				//if(buff[3] & 0x02) AsbStatus |=PRINTER_ABS_WAITINGONINSERT;
				//if(buff[3] & 0x08) AsbStatus |=PRINTER_ABS_WAITINGONEJECT;
				return 0;
				
			}
			else if(type==0){
				bAbsWatcher=false;
				buff[0]=0x1D;
				buff[1]=0x61;
				buff[2]=0x00;
				Output(3, buff);
				if(ThreadWaiting){
					pthread_cancel(asbthread);
			   }
				buff[0]=0x1D;
				buff[1]=0x28;
				buff[2]=0x40;
				buff[3]=2;
				buff[4]=0;
				buff[5]=97;
				buff[6]=0;
				Output(7, buff);
				return 0;
			}
			
		}
		return status;
	}

	// real time	
   if(type & PRINTER_GETSTATUS_PRINTER){
  	   Output(3, buff);
     	if(Input(1, &ret, false) <=0) return 0;
      if(ret & 0x04) status |=PRINTER_STATUS_KICK;
  	   if(ret & 0x08) status |=PRINTER_STATUS_ONLINE;
     	if(ret & 0x20) status |=PRINTER_STATUS_WOLRECOVERY;
      if(ret & 0x40) status |=PRINTER_STATUS_PAPERFEED;
  	}
   if(type & PRINTER_GETSTATUS_OFFLINE){
  	   buff[2]=0x02;
     	Output(3, buff);
      if(Input(1, &ret, false) <=0) return 0;
  	   if(ret & 0x04) status |=PRINTER_STATUS_COVER;
     	if(ret & 0x08) status |=PRINTER_STATUS_OLPAPERFEED;
      if(ret & 0x20) status |=PRINTER_STATUS_PAPEREND;
  	   if(ret & 0x40) status |=PRINTER_STATUS_ERROR;
  	}
  	if(type & PRINTER_GETSTATUS_ERROR){
     	buff[2]=0x03;
      Output(3, buff);
  	   if(Input(1, &ret, false) <=0) return 0;
     	if(ret & 0x04) status |=PRINTER_STATUS_MECHERROR;
      if(ret & 0x08) status |=PRINTER_STATUS_ACUTTERERROR;
  	   if(ret & 0x20) status |=PRINTER_STATUS_UNRECERROR;
     	if(ret & 0x40) status |=PRINTER_STATUS_AUNRECERROR;
   }
  	if(type & PRINTER_GETSTATUS_PAPERROLL){
     	buff[2]=0x04;
      Output(3, buff);
  	   if(Input(1, &ret, false) <=0) return 0;
     	if(ret & 0x0C) status |=PRINTER_STATUS_NEAREND;
      if(ret & 0x60) status |=PRINTER_STATUS_PAPERPRESENT;
  	}
   if(type & PRINTER_GETSTATUS_SLIP){
  	   buff[2]=0x05;
     	Output(3, buff);
      if(Input(1, &ret, false) <=0) return 0;
  	   if(!(ret & 0x02)) status |=PRINTER_STATUS_SLIPSELECT;
      if(ret & 0x08){
  	      status |=PRINTER_STATUS_SLIPWAIT;
     	   bWaitingPaperInsert=true;
     	}
     	if(!(ret & 0x20)) status |=PRINTER_STATUS_SLIPTOF;
     	if(!(ret & 0x40)) status |=PRINTER_STATUS_SLIPBOF;
  	}
  	if(type & PRINTER_GETSTATUS_VALIDATION){
     	buff[2]=0x06;
     	Output(3, buff);
     	if(Input(1, &ret, false) <=0) return 0;
     	if(ret & 0x04) status |=PRINTER_STATUS_VALSELECT;
     	if(ret & 0x08) status |=PRINTER_STATUS_VALWAIT;
      if(ret & 0x20) status |=PRINTER_STATUS_VALTOF;
  	   if(ret & 0x40) status |=PRINTER_STATUS_VALBOF;
  	}
  	if(type & PRINTER_GETSTATUS_MICR){
  	   buff[2]=0x08;
     	Output(4, buff);
      if(Input(1, &ret, false) <=0) return 0;
  	   if(!(ret & 0x04)) status |=PRINTER_STATUS_MICRSELECT;
     	if(ret & 0x08){
         status |=PRINTER_STATUS_MICRWAIT;
  	      bWaitingPaperInsert=true;
     	}
      if(!(ret & 0x20)) status |=PRINTER_STATUS_MICRTOF;
  	   if(!(ret & 0x40)) status |=PRINTER_STATUS_MICRBOF;
  	}
 
   return status;
}

int CPrinter::GetPaperStatus(BhPrintPaperType type, bool bPresent)
{
	unsigned char buff[8];
	unsigned char ret=0;
	
	switch(type){
	case PRN_PAPERTYPE_UNKNOWN:
		return 1;
   case PRN_PAPERTYPE_RECIEPT:
   case PRN_PAPERTYPE_SLIP:
   case PRN_PAPERTYPE_JOURNAL:
   case PRN_PAPERTYPE_ENDORSE:
   case PRN_PAPERTYPE_LABEL:
   case PRN_PAPERTYPE_MICR:
		break;
	case PRN_PAPERTYPE_CHEQUE:
		buff[0]=0x1B;      // enable sensors
	   buff[1]=0x63;
   	buff[2]=0x34;
      buff[3]=0x28;
      Output(4, (void *) buff);
			
		buff[0]=0x1B;      // select slip
	   buff[1]=0x63;
   	buff[2]=0x31;
      buff[3]=0x04;
      Output(4, (void *) buff);
		
      buff[0]=0x1D;
      buff[1]=0x28;
      buff[2]=0x47; 
      buff[3]=0x02; 
      buff[4]=0x00; 
      buff[5]=0x48; 
      buff[6]=0x04; //cut sheet control - select front of paper
      Output(7, buff);
		
		buff[0]=0x1D;
      buff[1]=0x72;
		buff[2]=3;
		break;
	}
	
	bReturn=0;
   
	while(1){
		if(bReturn){
			return 1;
		}
		Output(3, buff);
		Input(1, &ret, true);
		if(bPresent){
			if(ret==0){
				usleep(250000);
				continue;
			}
			else{
				return 0;
			}
		}
		else{
			if(ret==0){
				return 0;
			}
			else{
				usleep(250000);
				continue;
			}
		}
	}
	return 0;
}

void CPrinter::JustifyLine(int flags, char *s, char *e)
{
	if(!linebuffer) return;
	
	char nl=0x0A;
	memset(linebuffer, 0x20, reccharline);
	
	if(flags==0){
		memcpy(linebuffer, s, strlen(s));
		memcpy((linebuffer + (reccharline - strlen(e))), e, strlen(e));
		Output(reccharline, linebuffer);
	}
	else if(flags==1){
		memcpy((linebuffer + (reccharline - strlen(e))), e, strlen(e));
		Output(reccharline, linebuffer);
	}	
		
	Output(1, &nl);
}

int CPrinter::QuickParse(char *s)
{
   char *c=s;
   bool tok1, tok2, tok3;
   char bob[3];
   int  x;
   int  buffpos=0;
   char *buff=new char[strlen(s)+1];
   if(!buff) return -1;
   
   tok1=tok2=tok3=false;

   while(c){
      if(*c=='~') break;
      if(*c=='"' && !tok3){         
         tok3=true;
         c++;
         continue;
      }
      if(tok3){
         if(*c=='"'){         
            tok3=false;
            c++;
            continue;
         }
         else{
            buff[buffpos]=*c;
            buffpos++;
            c++;
            continue;
         }
      }
      if(*c=='0' && !tok1 && !tok2){   
         tok1=true;
         c++;
         continue;
      }
      if(*c=='x' && tok1){
         tok2=true;
         c++;
         continue;
      }
      if(tok1 && tok2){
         bob[0]=*c;
         c++;
         bob[1]=*c;
         bob[2]=0;
         x=atoi(bob);
         tok1=tok2=false;
         buff[buffpos]=(char) x;
         buffpos++;
         c++;
         continue;
      }
      c++;
   }
   s=(char *) realloc(s, buffpos);
   memcpy(s, buff, buffpos);
   delete buff;
   return buffpos;
}

////////////////////////////////////////////////////////
//
// for cheque writing
//
// If this not a page mode but 90 print rotation
// we need to build a two line char map with linespace
// setting the gap between the pay and amount line
//
//
void CPrinter::SpellMoney(double amount, char *result, char *majname, char *minname) 
{
   if(result==NULL) return;
   
   char buff[16];
   char cents[2]={'0', 0}; 
   char centst[2]={'0', 0};
   char ones[2]={'0', 0};
   char tens[2]={'0', 0};
   char hund[2]={'0', 0};
   char thou[2]={'0', 0};
   char tenth[2]={'0', 0};
   char hundth[2]={'0', 0};
   char *ptr;
   
   int len, dlen;
   
   if(amount > 999999.99) return;
   
   sprintf(buff, "%.2f", amount);
   len=strcspn(buff, ".");

   if(len > 0 ) ones[0] = buff[len];
   if(len > 1 ) tens[0] = buff[len-1];
   if(len > 2 ) hund[0] = buff[len-2];
   if(len > 3 ) thou[0] = buff[len-3];
   if(len > 4 ) tenth[0] = buff[len-4];
   if(len > 5 ) hundth[0] = buff[len-5];

   dlen=strlen(buff) - len;
   if(dlen > 0 ) cents[0] = buff[dlen];
   if(dlen > 1 ) centst[0] = buff[dlen+1];

   switch(atoi(hundth)){
   case 0: break;
   case 1: strcat(result, _("One Hundred ")); break;
   case 2: strcat(result, _("Two Hundred ")); break;
   case 3: strcat(result, _("Three Hundred ")); break;
   case 4: strcat(result, _("Four Hundred ")); break;
   case 5: strcat(result, _("Five Hundred ")); break;
   case 6: strcat(result, _("Six Hundred ")); break;
   case 7: strcat(result, _("Seven Hundred ")); break;
   case 8: strcat(result, _("Eight Hundred ")); break;
   case 9: strcat(result, _("Nine Hundred ")); break;
   }
   switch(atoi(tenth)){
   case 0: break;
   case 1: 
      switch(atoi(thou)){
         case 0: strcat(result, _("and Ten Thousand ")); thou[0]='0'; break;
         case 1: strcat(result, _("and Eleven Thousand ")); thou[0]='0'; break;
         case 2: strcat(result, _("and Twelve Thousand ")); thou[0]='0'; break;
         case 3: strcat(result, _("and Thirteen Thousand ")); thou[0]='0'; break;
         case 4: strcat(result, _("and Fourteen Thousand ")); thou[0]='0'; break;
         case 5: strcat(result, _("and Fifteen Thousand ")); thou[0]='0'; break;
         case 6: strcat(result, _("and Sixteen Thousand ")); thou[0]='0'; break;
         case 7: strcat(result, _("and Seventeen Thousand ")); thou[0]='0'; break;
         case 8: strcat(result, _("and Eighteen Thousand ")); thou[0]='0'; break;
         case 9: strcat(result, _("and Ninteen Thousand ")); thou[0]='0'; break;
      }
      break;
   case 2: strcat(result, _("and Twenty ")); break;
   case 3: strcat(result, _("and Thirty ")); break;
   case 4: strcat(result, _("and Forty ")); break;
   case 5: strcat(result, _("and Fifty ")); break;
   case 6: strcat(result, _("and Sixy ")); break;
   case 7: strcat(result, _("and Seventy ")); break;
   case 8: strcat(result, _("and Eightty ")); break;
   case 9: strcat(result, _("and Ninety ")); break;
   }
   
   switch(atoi(thou)){
   case 0: break;
   case 1: strcat(result, _("One Thousand ")); break;
   case 2: strcat(result, _("Two Thousand ")); break;
   case 3: strcat(result, _("Three Thousand ")); break;
   case 4: strcat(result, _("Four Thousand ")); break;
   case 5: strcat(result, _("Five Thousand ")); break;
   case 6: strcat(result, _("Six Thousand ")); break;
   case 7: strcat(result, _("Seven Thousand ")); break;
   case 8: strcat(result, _("Eight Thousand ")); break;
   case 9: strcat(result, _("Nine Thousand ")); break;
   }

   switch(atoi(hund)){
   case 0: break;
   case 1: 
      switch(atoi(tens)){
         case 0: strcat(result, _("Ten ")); break;
         case 1: strcat(result, _("Eleven ")); break;
         case 2: strcat(result, _("Twelve ")); break;
         case 3: strcat(result, _("Thirteen ")); break;
         case 4: strcat(result, _("Fourteen ")); break;
         case 5: strcat(result, _("Fifteen ")); break;
         case 6: strcat(result, _("Sixteen ")); break;
         case 7: strcat(result, _("Seventeen ")); break;
         case 8: strcat(result, _("Eighteen ")); break;
         case 9: strcat(result, _("Ninteen ")); break;
      }
      break;
   case 2: strcat(result, _("Two Hundred ")); break;
   case 3: strcat(result, _("Three Hundred ")); break;
   case 4: strcat(result, _("Four Hundred ")); break;
   case 5: strcat(result, _("Five Hundred ")); break;
   case 6: strcat(result, _("Six Hundred ")); break;
   case 7: strcat(result, _("Seven Hundred ")); break;
   case 8: strcat(result, _("Eight Hundred ")); break;
   case 9: strcat(result, _("Nine Hundred ")); break;
      break;
   }

   switch(atoi(tens)){
   case 0: break;
   case 1: 
      switch(atoi(thou)){
         case 0: strcat(result, _("and Ten")); thou[0]='0'; break;
         case 1: strcat(result, _("and Eleven")); thou[0]='0'; break;
         case 2: strcat(result, _("and Twelve")); thou[0]='0'; break;
         case 3: strcat(result, _("and Thirteen")); thou[0]='0'; break;
         case 4: strcat(result, _("and Fourteen")); thou[0]='0'; break;
         case 5: strcat(result, _("and Fifteen")); thou[0]='0'; break;
         case 6: strcat(result, _("and Sixteen")); thou[0]='0'; break;
         case 7: strcat(result, _("and Seventeen")); thou[0]='0'; break;
         case 8: strcat(result, _("and Eighteen")); thou[0]='0'; break;
         case 9: strcat(result, _("and Ninteen")); thou[0]='0'; break;
      }
      break;
   case 2: strcat(result, _("and Twenty ")); break;
   case 3: strcat(result, _("and Thirty ")); break;
   case 4: strcat(result, _("and Forty ")); break;
   case 5: strcat(result, _("and Fifty ")); break;
   case 6: strcat(result, _("and Sixy ")); break;
   case 7: strcat(result, _("and Seventy ")); break;
   case 8: strcat(result, _("and Eightty ")); break;
   case 9: strcat(result, _("and Ninety ")); break;
   }

   switch(atoi(ones)){
   case 0: break;
   case 1: strcat(result, _("One ")); strcat(result, majname); break;
   case 2: strcat(result, _("Two ")); strcat(result, majname); break;
   case 3: strcat(result, _("Three ")); strcat(result, majname); break;
   case 4: strcat(result, _("Four ")); strcat(result, majname); break;
   case 5: strcat(result, _("Five ")); strcat(result, majname); break;
   case 6: strcat(result, _("Six ")); strcat(result, majname); break;
   case 7: strcat(result, _("Seven ")); strcat(result, majname); break;
   case 8: strcat(result, _("Eight ")); strcat(result, majname); break;
   case 9: strcat(result, _("Nine ")); strcat(result, majname); break;
   }

   switch(atoi(centst)){
   case 0: break;
   case 1: 
      switch(atoi(cents)){
         case 0: strcat(result, _("and Ten")); cents[0]='0'; break;
         case 1: strcat(result, _("and Eleven")); cents[0]='0'; break;
         case 2: strcat(result, _("and Twelve")); cents[0]='0'; break;
         case 3: strcat(result, _("and Thirteen")); cents[0]='0'; break;
         case 4: strcat(result, _("and Fourteen")); cents[0]='0'; break;
         case 5: strcat(result, _("and Fifteen")); cents[0]='0'; break;
         case 6: strcat(result, _("and Sixteen")); cents[0]='0'; break;
         case 7: strcat(result, _("and Seventeen")); cents[0]='0'; break;
         case 8: strcat(result, _("and Eighteen")); cents[0]='0'; break;
         case 9: strcat(result, _("and Ninteen")); cents[0]='0'; break;
      }
      break;
   case 2: strcat(result, _("and Twenty ")); break;
   case 3: strcat(result, _("and Thirty ")); break;
   case 4: strcat(result, _("and Forty ")); break;
   case 5: strcat(result, _("and Fifty ")); break;
   case 6: strcat(result, _("and Sixy ")); break;
   case 7: strcat(result, _("and Seventy ")); break;
   case 8: strcat(result, _("and Eighty ")); break;
   case 9: strcat(result, _("and Ninety ")); break;
   }

   switch(atoi(cents)){
   case 0: break;
   case 1: strcat(result, _("One ")); strcat(result, minname); break;
   case 2: strcat(result, _("Two ")); strcat(result, minname); break;
   case 3: strcat(result, _("Three ")); strcat(result, minname); break;
   case 4: strcat(result, _("Four ")); strcat(result, minname); break;
   case 5: strcat(result, _("Five ")); strcat(result, minname); break;
   case 6: strcat(result, _("Six ")); strcat(result, minname); break;
   case 7: strcat(result, _("Seven ")); strcat(result, minname); break;
   case 8: strcat(result, _("Eight ")); strcat(result, minname); break;
   case 9: strcat(result, _("Nine ")); strcat(result, minname); break;
   }
}

void CPrinter::Signal(BhPrinterStatusEvent t, void *d)
{
   status_signal.emit(t, d);
}

void CPrinter::Error(BhPrinterErrorEvent t, void *d)
{
   error_signal.emit(t, d);
}

}
