#include <cstring>

#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <math.h>

#include "bhscales.h"


namespace bhpos{
   
CScales::CScales(BhDevType type) : CBhDevice(type)
{
   status=0;
	device=-1;
	
	bHighRes=false;
	getstatus=NULL;
	reset=NULL;
	zero=NULL;
	hiresweigh=NULL;
	weigh=NULL;
	expect=NULL;
}


CScales::~CScales()
{
}

int CScales::SetParameter(BhScalesParameter p, void *v)
{
	if(v==NULL) return 1;
	
   switch(p){
   case BHSCALES_PARAM_HIRES:
      bHighRes=*((bool *) v);
      break;
   case BHSCALES_PARAM_RESETSTRING:
      if(reset) delete reset;
      reset= new char [strlen((char *) v) + 1];
      if(reset) strcpy(reset, (char *) v);
      else return 2;
      break;
   case BHSCALES_PARAM_ZEROSTRING:
      if(zero) delete zero;
      zero= new char [strlen((char *) v) + 1];
      if(zero) strcpy(zero, (char *) v);
      else return 2;
      break;
   case BHSCALES_PARAM_WEIGHSTRING:
      if(weigh) delete weigh;
      weigh= new char [strlen((char *) v) + 1];
      if(weigh) strcpy(weigh, (char *) v);
      else return 2;
      break;
   case BHSCALES_PARAM_WEIGHHIRESSTRING:
      if(hiresweigh) delete hiresweigh;
      hiresweigh= new char [strlen((char *) v) + 1];
      if(hiresweigh) strcpy(hiresweigh, (char *) v);
      else return 2;
      break;
   case BHSCALES_PARAM_EXPECTSTRING:
      if(expect) delete expect;
      expect= new char [strlen((char *) v) + 1];
      if(expect) strcpy(expect, (char *) v);
      else return 2;
      break;
   }
   
   return 0;
}

int CScales::GetParameter(BhScalesParameter p, void *v)
{
  if(v==NULL) return 1;
	
   switch(p){
   case BHSCALES_PARAM_HIRES:
      *((bool *) v)=bHighRes;
      break;
   case BHSCALES_PARAM_RESETSTRING:
      if(reset) strcpy((char *) v, reset);
      else return 2;
      break;
   case BHSCALES_PARAM_ZEROSTRING:
      if(zero) strcpy((char *) v, zero);
      else return 2;
      break;
   case BHSCALES_PARAM_WEIGHSTRING:
      if(weigh) strcpy((char *) v, weigh);
      else return 2;
      break;
   case BHSCALES_PARAM_WEIGHHIRESSTRING:
      if(hiresweigh) strcpy((char *) v, hiresweigh);
      else return 2;
      break;
   case BHSCALES_PARAM_EXPECTSTRING:
      if(expect) strcpy((char *) v, expect);
      else return 2;
      break;
   }
   
   return 0;
}

bool CScales::SetDevice(int dev)
{
   device=dev;
    
   return false;
}

int CScales::GetStatus()
{
   char c;
   
   if(device==0) return -1;
   
   if(getstatus) write(device, getstatus, strlen(getstatus));
   read(device, &c, 1);
   status|=c;
   status&=0x0F;
   read(device, &c, 1);
   status|=(c << 4);
   status&=0xFF;
   
   return status;
}

int CScales::Weigh(bool hr)
{
   char buff[64];
   
   if(hr){
      if(!bHighRes) return 1;
      if(!hiresweigh) return 2;
      
      write(device, hiresweigh, strlen(hiresweigh));
      GetData(buff);
   }
   else{
      if(!weigh) return 2;
      
      write(device, weigh, strlen(weigh));
   }
   
   return 0;
}

int CScales::Zero()
{
   char c;
   
   if(device==0) return -1;
   if(zero) write(device, zero, strlen(zero));
   
   return GetStatus();
}

double CScales::Weight()
{
   char   buff[64];
   
   data=0.0;
   read(device, buff, 64);
   GetData(buff);
   
   return data;
}

//////////////////////////////////////////////////
//
//   NCI and ECR standard is: 
//   <LF>   -   start of data 
//   <CR>   -   field separator
//   <ETX>  -   end of data
//
//   Just the data format is different
//
//   ANSI Standard abbreviations for units are (2 letters):
//   lb, kg, oz, g, etc..(Note: ECR standard uses upper case)
//
int CScales::GetData(char *buff)
{
   if(!expect) return 1;

/* 
   int  num=0, i=1, x=0, pos=0;
   char temp[64];
   char weight[64];
   char *c=expect;
   
   //num=read(device, 64, temp);
   if(temp[0]!=0x0A) return 1;
   
   while(num){
		if(temp[i]!=0x03){ // end of the line
        continue; 
      }
      if(temp[i]==0x03){ // end of the line
         
      }
      if(temp[i]==0x0D){ // field sep
         pos=1;
      }
      if(temp[i]=='?'){ // unknown command
         
      }
      if(pos==0){
         if(weight[x]=='S'){
            status[x]=temp[i];
            pos=1;
         }
         weight[x]=temp[i];
         weight[x+1]=0;
      }
      else if(pos==1){
         
      }
      else if(pos==2){
      }
      x++;
      i++;
      c++;
      num--;
   }
   x=strlen(weight);
   u[0]=weight[x-1];
   u[1]=weight[x];
   u[3]=0;
   wieght[x-1]=0;
   *w=atof(weight);
*/   
}

}
